# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>
#from . import pattern_preview

import subprocess, os, sys, platform
bl_info = {
    "name": "POV-Ray 3.7",
    "author": "Campbell Barton, Silvio Falcinelli, Maurice Raybaud, Constantin Rahn, Bastien Montagne, Leonid Desyatkov",
    "version": (1,0,0),
    "blender": (2, 69, 0),
    "location": "Render > Engine > POV-Ray 3.7",
    "description": "Basic POV-Ray 3.7 integration for blender",
    "wiki_url": "http://wiki.blender.org/index.php/Extensions:2.6/Py/"
                "Scripts/Render/POV-Ray",
    "tracker_url": "https://projects.blender.org/tracker/index.php?"
                   "func=detail&aid=23145",
    "category": "Render"}

if "bpy" in locals():
    import imp
    imp.reload(ui)
    imp.reload(render)
    #imp.reload(update_files)

else:
    import bpy
    from bpy.types import (AddonPreferences,
                           PropertyGroup,
                           Operator)
    from bpy.props import (StringProperty,
                           BoolProperty,
                           IntProperty,
                           FloatProperty,
                           FloatVectorProperty,
                           EnumProperty,
                           PointerProperty,
                           CollectionProperty)
    from . import ui
    from . import render
from bpy_extras.io_utils import ExportHelper

def string_strip_hyphen(name):
    return name.replace("-", "")

def use_shader_nodes(self,context):
    try:
        mat=context.material
        n=0
        for g in bpy.data.node_groups:
            if g.name == mat.name:
                n+=1
        if n == 0:
            ntree=bpy.data.node_groups.new("%s"%mat.name,'PovrayTreeType') 
    except:
        pass
def call_mat_pattern_create(self,context):
    try:
        ob=context.object
        if ob.pov.mat_patterns == "cubic":
            povName = string_strip_hyphen(bpy.path.clean_name(ob.name))        
            colors=[(1,0,0),(1,1,0),(1,1,0.5),(0,1,0),(0.5,0.5,1),(0,0,1)]
            for i in range(0,6):
                bpy.ops.object.material_slot_add()
                bpy.ops.object.material_slot_assign()
                mat = bpy.data.materials.new("Mat_%s_%s"%(povName,i))
                mat.diffuse_color = colors[i][:]
                ob.material_slots[ob.material_slots.__len__() - 1].material = mat 
    except:
        pass

def mat_texslots_update(self, context):
    ob=context.object
    mat=context.material
    pigname=self.layer_pigment
    texuse=False
    for matslot in ob.material_slots:
        obmat=matslot.material
        if obmat:
            for texslot in obmat.texture_slots:
                if texslot:
                    tex=texslot.texture
                    if tex.name == pigname:
                        texuse = True
    if texuse == False:
        for tex in bpy.data.textures:
            if tex.name == pigname:
                deltex = mat.texture_slots.add()
                deltex.texture = tex

def call_update_scene_pigments(self, context):
    scene=context.scene
    tex=context.texture
    try:
        if tex.pov.tex_as_pigment:
            pig=scene.povray_pigments.add()
            pig.name=tex.name
            pig.index=len(scene.povray_pigments)-1
        else:
            for pig in scene.povray_pigments:
                if pig.name==tex.name:
                    scene.povray_pigments.remove(pig.index)
    except:
        pass

def call_update_scene_normals(self, context):
    scene=context.scene
    tex=context.texture
    try:
        if tex.pov.tex_as_normal:
            pig=scene.povray_normals.add()
            pig.name=tex.name
            pig.index=len(scene.povray_normals)-1
        else:
            for pig in scene.povray_normals:
                if pig.name==tex.name:
                    scene.povray_normals.remove(pig.index)
    except:
        pass

def call_update_scene_finishes(self, context):
    scene=context.scene
    mat=context.material
    try:
        if mat.pov.povray_finish:
            pig=scene.povray_finishes.add()
            pig.name=mat.name
            pig.index=len(scene.povray_finishes)-1
        else:
            for pig in scene.povray_finishes:
                if pig.name==mat.name:
                    scene.povray_finishes.remove(pig.index)
    except:
        pass

def call_update_scene_pigment_patterns(self, context):
    scene=context.scene
    tex=context.texture
    try:
        if tex.pov.tex_as_ppattern:
            pig=scene.povray_pigment_patterns.add()
            pig.name=tex.name
            pig.index=len(scene.povray_pigment_patterns)-1
        else:
            for pig in scene.povray_pigment_patterns:
                if pig.name==tex.name:
                    scene.povray_pigment_patterns.remove(pig.index)
    except:
        pass

def call_add_slope_map(self, context):
    try:
        tex=context.texture
        mat=context.material
        number=0
        for entry in tex.povray_slope:
            if entry:
                number+=1
        if number == 0:
            value=[0,0.5,0.5,1]
            height=[0,1,1,0]
            slope=[1,1,-1,-1]
            for i in range(0,4):
                entry=tex.povray_slope.add()
                entry.name="slope_entry.%s"%i
                entry.slope_value=value[i]
                entry.slope_height=height[i]
                entry.slope_slope=slope[i]
    except:
        pass 

def call_mat_spec_update(self, context):
    try:
        mat = context.material
        percent = mat.pov.spec_size * 100
        s_h = 511-(511 * percent / 100)
        if s_h<1:s_h=1 
        mat.specular_hardness=s_h
    except:
        pass 

def call_tex_type_update(self, context):
    try:
        tex = context.texture
        tex.type = "IMAGE"
        if tex is not None and tex.pov.tex_type == "PATTERN":
            tex.use_color_ramp = True
        else: tex.use_color_ramp = False 
    except:
        pass

def call_nor_type_update(self, context):
    try:
        tex = context.texture
        tex.type = "IMAGE"
        if tex is not None and tex.pov.nor_type == "PATTERN":
            tex.use_color_ramp = True
        else: tex.use_color_ramp = False 
    except:
        pass

def call_tex_image_open(self, context): #####################откорректировать для винды################################
    try:
        tex = context.texture
        tex.use_nodes = True
        tree = tex.node_tree
        links = tree.links
        for n in tree.nodes:
            tree.nodes.remove(n)
        im = tree.nodes.new("TextureNodeImage")  
        im.image = bpy.data.images.load(tex.pov.tex_image_path)
        name=tex.pov.tex_image_path
        name=name.split("/")
        name=name[len(name)-1]
        im.name = name  
        im.location = 200,200
        previewer = tree.nodes.new('TextureNodeOutput')    
        previewer.label = "Preview"
        previewer.location = 400,400
        links.new(im.outputs[0],previewer.inputs[0]) 
        tex.pov.tex_images_data=name        
    except:
        pass 

def call_tex_image_link(self, context):
    try:
        tex = context.texture
        name=tex.pov.tex_image_path
        name=name.split("/")
        name=name[len(name)-1]
        for im in bpy.data.images:
            n=0
            if im.name == tex.pov.tex_images_data and im.name != name:
                tex.pov.tex_image_path = im.filepath

    except:
        pass 

def call_m_transparent(self, context):
    mat = context.material
    try:
        if mat.pov.m_transparent:
            mat.use_transparency=True
            mat.transparency_method = 'RAYTRACE'
            mat.raytrace_transparency.fresnel=mat.pov.m_filter
            mat.raytrace_transparency.fresnel_factor=5.0
            mat.raytrace_transparency.filter = 1.0
        if mat.pov.m_transparent==False:
            mat.raytrace_transparency.fresnel_factor=0.0
            mat.raytrace_transparency.fresnel=0.0
            mat.use_transparency=False
    except:
        pass

def call_mat_transmit(self, context):
    mat = context.material
    try:
        mat.alpha = 1-mat.pov.m_transmit
        mat.raytrace_transparency.filter = 1-mat.pov.m_transmit
    except:
        pass

def call_mat_filter(self, context):
    mat = context.material
    try:
        mat.raytrace_transparency.fresnel = mat.pov.m_filter
    except:
        pass

def call_obj_ior(self, context):
    ob = context.object
    try:
        slots=ob.material_slots
        for sl in slots:
            if sl.material.use_transparency and sl.material.transparency_method == 'RAYTRACE':
                sl.material.raytrace_transparency.ior=ob.pov.m_ior
    except:
        pass


def call_cr_gradient_add(self, context): 
    try:
        ob=context.object
        if obj.pov.cr_fill_stroke in {"FILLGRADIENT","STROKEGRADIENT"}:
            bpy.context.scene.use_nodes = True
            tree = bpy.context.scene.node_tree
            obNodeName=ob.name+"CairoColor"
            obNode=0
            for n in tree.nodes:
                if n and n.name == obNodeName:
                    obNode=1
            if obNode==0:
                crNode = tree.nodes.new('CompositorNodeValToRGB')      
                crNode.location = 0,200
                crNode.name = obNodeName
    except:
        pass

def call_fog_use(self, context):
    try:
        world=context.world
        if len(world.povray_fogs) == 0:
            layer=world.povray_fogs.add()
            layer.index=0
    except:
        pass

def call_start_preview(self, context):
    pass
################ if material.use_transparency and material.transparency_method == 'RAYTRACE':
###############################################################################
# Lamp POV properties.
###############################################################################
class RenderPovSettingsLamp(PropertyGroup):
    # Area Lamp Options

    axis_1_vector_x = FloatProperty(
            name="X", description="pov_axis_1_vector_x",
            min=0.0, max=1.0, default=0.0)

    axis_1_vector_y = FloatProperty(
            name="Y", description="pov_axis_1_vector_y",
            min=0.0, max=1.0, default=0.0)           

    axis_1_vector_z = FloatProperty(
            name="Z", description="pov_axis_1_vector_z",
            min=0.0, max=1.0, default=0.0)

    axis_2_vector_x = FloatProperty(
            name="X", description="pov_axis_2_vector_x",
            min=0.0, max=1.0, default=0.0)

    axis_2_vector_y = FloatProperty(
            name="Y", description="pov_axis_2_vector_y",
            min=0.0, max=1.0, default=0.0)

    axis_2_vector_z = FloatProperty(
            name="Z", description="pov_axis_2_vector_z",
            min=0.0, max=1.0, default=0.0)

    area_size_1 = FloatProperty(
            name="Y",
            description="Area size 1. Float(0...10)!")

    area_size_2 = FloatProperty(
            name="Z",
            description="Area size 2. Float(0...10)!")

    media_interaction = BoolProperty(
            name="Media interaction",
            description="Media interaction",
            default=True)

    media_attenuation = BoolProperty(
            name="Media attenuation",
            description="Media attenuation",
            default=False)

    falloff_spot = FloatProperty(
            name="Z", description="Fade distance",
            min=0.0, max=100.0, default=0.0)
            
    looks_like = BoolProperty(
            name="Looks like",
            description="Object as lamp",
            default=False)           
###############################################################################
# Scene POV properties.
###############################################################################
class RenderPovSettingsScene(PropertyGroup):
    # File Options
    tempfiles_enable = BoolProperty(
            name="Enable Tempfiles",
            description="Enable the OS-Tempfiles. Otherwise set the path where to save the files",
            default=False)
    deletefiles_enable = BoolProperty(
            name="Delete files",
            description="Delete files after rendering. Doesn't work with the image",
            default=False)
    scene_name = StringProperty(
            name="Scene Name",
            description="Name of POV-Ray scene to create. Empty name will use the name of "
                        "the blend file",
            maxlen=1024)
    scene_path = StringProperty(
            name="Export scene path",
            # description="Path to directory where the exported scene (POV and INI) is created",  # Bug in POV-Ray RC3
            description="Path to directory where the files are created",
            maxlen=1024, subtype="DIR_PATH")
    renderimage_path = StringProperty(
            name="Rendered image path",
            description="Full path to directory where the rendered image is saved",
            maxlen=1024, subtype="DIR_PATH")
    list_lf_enable = BoolProperty(
            name="LF in lists",
            description="Enable line breaks in lists (vectors and indices). Disabled: "
                        "lists are exported in one line",
            default=True)

    # Not a real pov option, just to know if we should write
    radio_enable = BoolProperty(
            name="Enable Radiosity",
            description="Enable POV-Rays radiosity calculation",
            default=False)
    radio_display_advanced = BoolProperty(
            name="Advanced Options",
            description="Show advanced options",
            default=False)
    media_enable = BoolProperty(
            name="Enable Media",
            description="Enable POV-Rays atmospheric media",
            default=False)
    media_samples = IntProperty(
            name="Samples",
            description="Number of samples taken from camera to first object "
                        "encountered along ray path for media calculation",
            min=1, max=100, default=35)

    media_color = FloatVectorProperty(
            name="Media Color", description="The atmospheric media color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.001, 0.001, 0.001), options={'ANIMATABLE'}, subtype='COLOR')

    baking_enable = BoolProperty(
            name="Enable Baking",
            description="Enable POV-Rays texture baking",
            default=False)
    indentation_character = EnumProperty(
            name="Indentation",
            description="Select the indentation type",
            items=(('NONE', "None", "No indentation"),
                   ('TAB', "Tabs", "Indentation with tabs"),
                   ('SPACE', "Spaces", "Indentation with spaces")),
            default='SPACE')
    indentation_spaces = IntProperty(
            name="Quantity of spaces",
            description="The number of spaces for indentation",
            min=1, max=10, default=4)

    comments_enable = BoolProperty(
            name="Enable Comments",
            description="Add comments to pov file",
            default=True)

    # Real pov options
    command_line_switches = StringProperty(
            name="Command Line Switches",
            description="Command line switches consist of a + (plus) or - (minus) sign, followed "
                        "by one or more alphabetic characters and possibly a numeric value",
            maxlen=500)

    antialias_enable = BoolProperty(
            name="Anti-Alias", description="Enable Anti-Aliasing",
            default=True)

    antialias_method = EnumProperty(
            name="Method",
            description="AA-sampling method. Type 1 is an adaptive, non-recursive, super-sampling "
                        "method. Type 2 is an adaptive and recursive super-sampling method",
            items=(("1", "non-recursive AA", "Type 1 Sampling in POV-Ray"),
                   ("2", "recursive AA", "Type 2 Sampling in POV-Ray"),
                   ("3", "adaptive AA", "Type 3 Sampling in POV-Ray")),
            default="3")

    antialias_depth = IntProperty(
            name="Antialias Depth", description="Depth of pixel for sampling",
            min=1, max=9, default=3)

    antialias_threshold = FloatProperty(
            name="Antialias Threshold", description="Tolerance for sub-pixels",
            min=0.0, max=1.0, soft_min=0.05, soft_max=0.5, default=0.1)

    jitter_enable = BoolProperty(
            name="Jitter",
            description="Enable Jittering. Adds noise into the sampling process (it should be "
                        "avoided to use jitter in animation)",
            default=False)

    jitter_amount = FloatProperty(
            name="Jitter Amount", description="Amount of jittering",
            min=0.0, max=1.0, soft_min=0.01, soft_max=1.0, default=0.0)

    antialias_gamma = FloatProperty(
            name="Antialias Gamma",
            description="POV-Ray compares gamma-adjusted values for super sampling. Antialias "
                        "Gamma sets the Gamma before comparison",
            min=0.0, max=5.0, soft_min=0.01, soft_max=2.5, default=2.5)

    assumed_gamma = FloatProperty(
            name="Assumed gamma",
            description="Assumed gamma",
            min=0.0, max=10.0, default=1.0)

    max_trace_level = IntProperty(
            name="Max Trace Level",
            description="Number of reflections/refractions allowed on ray path",
            min=0, max=256, default=3)

    display_gamma = FloatProperty(
            name="Display gamma",
            description="Display gamma",
            min=0.0, max=10.0, default=1.0)

    mm_per_unit = FloatProperty(
            name="mm_per_unit",
            description="mm_per_unit",
            min=0.0, max=1000.0, default=10.0)

    sss_enable = BoolProperty(
            name="SSS enable",
            description="Enable Subsurface Light Transport",
            default=False)

    sss_samples_enable = BoolProperty(
            name="SSS samples enable",
            description="Enable Sampling",
            default=False)

    sss_radiosity = BoolProperty(
            name="SSS radiosity",
            description="",
            default=False)

    sss_samples_max = IntProperty(
            name="Samples max",
            description="",
            min=1, max=1000, default=400)

    sss_samples_min = IntProperty(
            name="Samples min",
            description="",
            min=1, max=1000, default=40)

    photon_enable = BoolProperty(
            name="Photons",
            description="Enable global photons",
            default=False)

    photon_count = IntProperty(
            name="Photons count",
            description="Photons count",
            min=1, max=100000000, default=20000)

    photon_max_trace_level = IntProperty(
            name="Max Trace Level",
            description="Number of reflections/refractions allowed on ray path",
            min=1, max=256, default=5)

    photon_media_enable = BoolProperty(
            name="Photons",
            description="Enable global photons",
            default=False)

    photon_media_steps = IntProperty(
            name="Photons count",
            description="Photons count",
            min=1, max=1000, default=100)

    photon_media_factor = FloatProperty(
            name="Factor",
            description="Photon media factor",
            min=1.0, max=10.0, default=1.0)

    photon_advanced = BoolProperty(
            name="Advanced",
            description="Advanced options",
            default=False)

    photon_adc = BoolProperty(
            name="Photon ADC",
            description="Photon ADC enable",
            default=False)

    photon_adc_bailout = FloatProperty(
            name="ADC_bailout",
            description="The adc_bailout for photons. Use adc_bailout = "
                        "0.01 / brightest_ambient_object for good results",
            min=0.0, max=1000.0, soft_min=0.0, soft_max=1.0, default=0.1, precision=3)

    photon_gather = BoolProperty(
            name="Gather",
            description="Photon gather",
            default=False)

    photon_gather_min = IntProperty(
            name="Gather Min", description="Minimum number of photons gathered for each point",
            min=1, max=256, default=20)

    photon_gather_max = IntProperty(
            name="Gather Max", description="Maximum number of photons gathered for each point",
            min=1, max=256, default=100)

    photon_autostop = BoolProperty(
            name="Autostop",
            description="Photon autostop",
            default=False)

    photon_autostop_value = FloatProperty(
            name="Value",
            description="Photon autostop value",
            min=0.0, max=1.0, default=0.0)

    photon_savefile = BoolProperty(
            name="Save",
            description="Photon map file save",
            default=False)

    photon_openfile = BoolProperty(
            name="Open",
            description="Photon map file open",
            default=False)
    ########################## radio ###############################################
    radio_adc_bailout = FloatProperty(
            name="ADC Bailout",
            description="The adc_bailout for radiosity rays. Use "
                        "adc_bailout = 0.01 / brightest_ambient_object for good results",
            min=0.0, max=1000.0, soft_min=0.0, soft_max=1.0, default=0.01, precision=3)

    radio_always_sample = BoolProperty(
            name="Always Sample",
            description="Only use the data from the pretrace step and not gather "
                        "any new samples during the final radiosity pass",
            default=True)

    radio_brightness = FloatProperty(
            name="Brightness",
            description="Amount objects are brightened before being returned "
                        "upwards to the rest of the system",
            min=0.0, max=1000.0, soft_min=0.0, soft_max=10.0, default=1.0)

    radio_count = IntProperty(
            name="Ray Count",
            description="Number of rays for each new radiosity value to be calculated "
                        "(halton sequence over 1600)",
            min=1, max=10000, soft_max=1600, default=35)

    radio_error_bound = FloatProperty(
            name="Error Bound",
            description="One of the two main speed/quality tuning values, "
                        "lower values are more accurate",
            min=0.0, max=1000.0, soft_min=0.1, soft_max=10.0, default=1.8)

    radio_gray_threshold = FloatProperty(
            name="Gray Threshold",
            description="One of the two main speed/quality tuning values, "
                        "lower values are more accurate",
            min=0.0, max=1.0, soft_min=0, soft_max=1, default=0.0)

    radio_low_error_factor = FloatProperty(
            name="Low Error Factor",
            description="Just enough samples is slightly blotchy. Low error changes error "
                        "tolerance for less critical last refining pass",
            min=0.0, max=1.0, soft_min=0.0, soft_max=1.0, default=0.5)

    # max_sample - not available yet
    radio_media = BoolProperty(
            name="Media", description="Radiosity estimation can be affected by media",
            default=False)

    radio_minimum_reuse = FloatProperty(
            name="Minimum Reuse",
            description="Fraction of the screen width which sets the minimum radius of reuse "
                        "for each sample point (At values higher than 2% expect errors)",
            min=0.0, max=1.0, soft_min=0.1, soft_max=0.1, default=0.015, precision=3)

    radio_nearest_count = IntProperty(
            name="Nearest Count",
            description="Number of old ambient values blended together to "
                        "create a new interpolated value",
            min=1, max=20, default=5)

    radio_normal = BoolProperty(
            name="Normals", description="Radiosity estimation can be affected by normals",
            default=False)

    radio_recursion_limit = IntProperty(
            name="Recursion Limit",
            description="how many recursion levels are used to calculate "
                        "the diffuse inter-reflection",
            min=1, max=20, default=3)

    radio_pretrace_start = FloatProperty(
            name="Pretrace Start",
            description="Fraction of the screen width which sets the size of the "
                        "blocks in the mosaic preview first pass",
            min=0.01, max=1.00, soft_min=0.02, soft_max=1.0, default=0.08)

    radio_pretrace_end = FloatProperty(
            name="Pretrace End",
            description="Fraction of the screen width which sets the size of the blocks "
                        "in the mosaic preview last pass",
            min=0.001, max=1.00, soft_min=0.01, soft_max=1.00, default=0.04, precision=3)
    ################### pycairo python version #########################################
    pycairo_pyversion = EnumProperty(
            name="Pycairo python version",
            description="Pycairo python version",
            items=(('PYTHON2', "Python 2", ""),('PYTHON3', "Python 3", "")),
            default='PYTHON2')
###############################################################################
# Material POV properties.
###############################################################################
class RenderPovSettingsMaterial(PropertyGroup):
    ############## finish ###################
    povray_finish = BoolProperty(
            name="Finish",
            description="",
            update=call_update_scene_finishes,
            default=False)

    finish_as_mat = BoolProperty(
            name="Use as material",
            description="",
            default=True)

    m_use_ambient = BoolProperty(
            name="Ambient",
            description="Use ambient",
            default=False)

    m_ambient_color = FloatVectorProperty(
            name="Ambient color", description="Ambient color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.0, 0.0, 0.0), options={'ANIMATABLE'}, subtype='COLOR')

    m_use_diffuse = BoolProperty(
            name="Diffuse",
            description="Use diffuse",
            default=True)

    m_diffuse_albedo = BoolProperty(
            name="Albedo",
            description="Use albedo",
            default=False)

    m_use_emission = BoolProperty(
            name="Emission",
            description="Use emission",
            default=False)

    m_use_brilliance = BoolProperty(
            name="Brilliance",
            description="Use brilliance",
            default=False)

    m_brilliance = FloatProperty(
            name="Brilliance",
            description="Brilliance",
            min=0.0, max=10.0, default=2.0)

    m_use_spec = BoolProperty(
            name="Spec",
            description="Use spec",
            default=True)

    spec_size = FloatProperty(
            name="Spec size",
            description="Specular size",
            update=call_mat_spec_update,
            min=0.0001, max=1.0, default=0.01) 

    spec_use_phong = BoolProperty(
            name="Phong",
            description="Phong",
            default=False)

    spec_use_albedo = BoolProperty(
            name="Albedo",
            description="Albedo",
            default=False)

    m_use_specmetallic = BoolProperty(
            name="Specular metallic",
            description="Use specular metallic",
            default=False)

    m_specmetallic = FloatProperty(
            name="Value",
            description="Specular metallic value",
            min=0.0, max=1.0, default=0.0)

    m_use_crand = BoolProperty(
            name="Crand",
            description="Albedo",
            default=False)

    m_crand = FloatProperty(
            name="Graininess",
            description="Crand Graininess",
            min=0.0, max=1.0, default=0.0)

    m_metallic = FloatProperty(
            name="Metallic",
            description="Metallic",
            min=0.0, max=1.0, default=0.0)

    m_reflect_fresnel = BoolProperty(
            name="Fresnel reflect",
            description="Fresnel reflect",
            default=False)

    m_reflect_falloff = FloatProperty(
            name="Reflect falloff",
            description="Reflect falloff",
            min=0.0, max=10.0, default=0.0)

    m_reflect_exp = FloatProperty(
            name="Reflect exponent",
            description="Reflect exponent",
            min=0.0, max=10.0, default=0.0)

    irid_enable = BoolProperty(
            name="Enable Iridescence",
            description="Newton's thin film interference (like an oil slick on a puddle of "
                        "water or the rainbow hues of a soap bubble.)",
            default=False)

    conserve_energy = BoolProperty(
            name="Conserve Energy",
            description="Light transmitted is more correctly reduced by mirror reflections, "
                        "also the sum of diffuse and translucency gets reduced below one ",
            default=True)

    irid_amount = FloatProperty(
            name="amount",
            description="Contribution of the iridescence effect to the overall surface color. "
                        "As a rule of thumb keep to around 0.25 (25% contribution) or less, "
                        "but experiment. If the surface is coming out too white, try lowering "
                        "the diffuse and possibly the ambient values of the surface",
            min=0.0, max=1.0, soft_min=0.01, soft_max=1.0, default=0.25)

    irid_thickness = FloatProperty(
            name="thickness",
            description="A very thin film will have a high frequency of color changes while a "
                        "thick film will have large areas of color",
            min=0.0, max=1000.0, soft_min=0.1, soft_max=10.0, default=1)

    irid_turbulence = FloatProperty(
            name="turbulence", description="This parameter varies the thickness",
            min=0.0, max=10.0, soft_min=0.000, soft_max=1.0, default=0)

    sss_mat_enable = BoolProperty(
            name="sss_mat_enable",
            description="",
            default=False)

    sss_translucency_color = FloatVectorProperty(
            name="Translucency color", description="Subsurface light transport translucency color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.0, 0.0, 0.0), options={'ANIMATABLE'}, subtype='COLOR')

    ################# POVMEDIA ###################################################################
    media_ob_enable = BoolProperty(
            name="Use as Povray media",
            description="Present object for effects",
            default=False)

    absorption_media_ob = BoolProperty(
            name="Media Absorption",
            description="Media Absorption",
            default=False)

    absorption_val = FloatProperty(
            name="Value",
            description="Absorption value",
            min=0.0, max=50.0, default=0.0)

    emission_media_ob = BoolProperty(
            name="Media Emission",
            description="Media Emission",
            default=False)

    emission_val = FloatProperty(
            name="Value",
            description="Emission value",
            min=0.0, max=5.0, default=0.0)

    scattering_media_ob = BoolProperty(
            name="Media Scattering",
            description="Media Scattering",
            default=False)

    s_media_type = IntProperty(
            name="Scattering type",
            description="1 ... 5",
            min=1, max=5, default=1)

    s_media_color = FloatVectorProperty(
            name="Scattering Media Color", description="The Scattering media color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.001, 0.001, 0.001), options={'ANIMATABLE'}, subtype='COLOR')

    ext_s_media = FloatProperty(
            name="Extinction",
            description="Extinction",
            min=-10.0, max=10.0, default=0.0)

    ecc_s_media = FloatProperty(
            name="Eccentricity",
            description="Eccentricity",
            min=-1.0, max=1.0, default=0.0)

    s_samples = IntProperty(
            name="Scattering samples",
            description="1 ... 256",
            min=1, max=256, default=30)
    ############## END POVMEDIA ###############################################################
    ############## Material texture ###############################################################
    povray_mat_type = EnumProperty(
            name="Type",
            description="Texture type",
            items=(('DEFAULT', "Default", ""),('LAYERED', "Layered", ""),
            ('PATTERN', "Pattern", ""),('PIGMENTPATTERN', "Pigment pattern", "")),
            default='DEFAULT')

    m_layer_number = IntProperty(
            name="Layer",
            description="1 ... 256",
            min=0, max=256, default=0)

    m_transparent = BoolProperty(
            name="Use transparency",
            description="Use transparency",
            update=call_m_transparent,
            default=False)

    m_transmit = FloatProperty(
            name="Transmit",
            description="Transmit",
            update=call_mat_transmit,
            min=0.0, max=1.0, default=0.0)            

    m_filter = FloatProperty(
            name="Filter",
            description="Filter",
            update=call_mat_filter,
            min=0.0, max=1.0, default=0.0)

    material_use_nodes = BoolProperty(
            name="Use nodes",
            description="Use node",
            update=use_shader_nodes,
            default=False)

    m_pp = StringProperty(
            name="Pigment Pattern",
            description="")

    m_pp_first = StringProperty(
            name="First",
            description="First material")

    m_pp_second = StringProperty(
            name="Second",
            description="Second material")

    m_first_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=1.0, default=0.0)

    m_second_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=1.0, default=1.0)
###############################################################################
# Texture POV properties.
###############################################################################
class RenderPovSettingsTexture(PropertyGroup):

    tex_type = EnumProperty(
            name="Texture type",
            description="Texture type",
            items=(('NONE', "None", ""),('IMAGE', "Image", ""),('PATTERN', "Pattern", "")),
            update=call_tex_type_update,
            default='NONE')

    nor_type = EnumProperty(
            name="Normal type",
            description="Normal type",
            items=(('NONE', "None", ""),('IMAGE', "Image", ""),('PATTERN', "Pattern", "")),
            update=call_nor_type_update,
            default='NONE')

    tex_image_path = StringProperty(
            name="Path image",
            description="Path image",
            update=call_tex_image_open,
            maxlen=1024, subtype="FILE_PATH") 

    tex_images_data = StringProperty(
            name="Image from data:",
            update=call_tex_image_link,
            description="Select image in blend_data images")

    tex_coords = EnumProperty(
            name="Texture coords",
            description="Texture coords",
            items=(('UV', "UV", ""),('GENERATED', "Generated", "")),                   
            default='UV')

    tex_as_pigment = BoolProperty(
            name="Pigment",
            description="",
            update=call_update_scene_pigments,
            default=False)

    tex_pigment_color = FloatVectorProperty(
            name="Pigment color", description="Pigment color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(1.0, 1.0, 1.0), options={'ANIMATABLE'}, subtype='COLOR')

    tex_pigment_filter = FloatProperty(
            name="Filter", description="",
            min=0.0, max=1.0, default=0)

    tex_pigment_transmit = FloatProperty(
            name="Transmit", description="",
            min=0.0, max=1.0, default=0)

    tex_as_material = BoolProperty(
            name="Material",
            description="Use as material",
            default=False)

    tex_ip_premul = BoolProperty(
            name="Premultiplied",
            description="",
            default=False)

    tex_ip_alpha = BoolProperty(
            name="Use alpha",
            description="",
            default=False)

    tex_ip_gamma = FloatProperty(
            name="Gamma",
            description="",
            min=0.1, max=10.00, default=0.1)

    tex_as_normal = BoolProperty(
            name="Normal",
            description="Use as normal map",
            update=call_update_scene_normals,
            default=False)

    tex_normal_gray = FloatProperty(
            name="Value",
            description="Gray value",
            min=0.0, max=1.0,default=1.0)

    tex_as_ppattern = BoolProperty(
            name="Pigment pattern",
            description="Use as pigment pattern",
            update=call_update_scene_pigment_patterns,
            default=False)

    tex_use_pmap = BoolProperty(
            name="Pigment map",
            description="Use pigment map",
            default=False)

    pigment_entry_number = IntProperty(
            name="Remove entry N",
            description="Select entry for remove",
            min=0,max=255,default=0)

    tex_use_slope = BoolProperty(
            name="Use slope map",
            description="",
            update=call_add_slope_map,
            default=False)

    tex_slope_accuracy = FloatProperty(
            name="Accuracy",
            description="",
            min=0.0001, max=1.0, default=0.02) 

    slope_entry_number = IntProperty(
            name="Remove entry N",
            description="Select entry for remove",
            min=0,max=255,default=0)

    tex_nor_color_map = BoolProperty(
            name="Use color map",
            description="Use color map",
            default=False)

    tex_nor = FloatProperty(
            name="Normal",
            description="",
            min=0.0, max=5.0, default=0.0) 

    tex_use_absorp = BoolProperty(
            name="Use absorption",
            description="Use as absorption density",
            default=False)

    tex_absorp = FloatProperty(
            name="Absorption",
            description="",
            min=0.0, max=1.0, default=1.0) 

    tex_use_emis = BoolProperty(
            name="Use emission",
            description="Use as emission density",
            default=False)

    tex_emis = FloatProperty(
            name="Emission",
            description="",
            min=0.0, max=1.0, default=1.0) 

    tex_rot_x = FloatProperty(
            name="Rotate X",
            description="",
            min=-180.0, max=180.0, default=0.0) 

    tex_rot_y = FloatProperty(
            name="Rotate Y",
            description="",
            min=-180.0, max=180.0, default=0.0) 

    tex_rot_z = FloatProperty(
            name="Rotate Z",
            description="",
            min=-180.0, max=180.0, default=0.0) 
            
    tex_mov_x = FloatProperty(
            name="Move X",
            description="",
            min=-100000.0, max=100000.0, default=0.0) 

    tex_mov_y = FloatProperty(
            name="Move Y",
            description="",
            min=-100000.0, max=100000.0, default=0.0) 

    tex_mov_z = FloatProperty(
            name="Move Z",
            description="",
            min=-100000.0, max=100000.0, default=0.0)             

    tex_scale_x = FloatProperty(
            name="Scale X",
            description="",
            min=-10000.0, max=10000.0, default=1.0) 

    tex_scale_y = FloatProperty(
            name="Scale Y",
            description="",
            min=-10000.0, max=10000.0, default=1.0) 

    tex_scale_z = FloatProperty(
            name="Scale Z",
            description="",
            min=-10000.0, max=10000.0, default=1.0) 
    #################################################################################################################LH


    colored_patterns = EnumProperty(
            name="Colored patterns",
            description="Select the colored pattern",
            items=(('agate', "Agate", ""), ('boxed', "Boxed", ""), ('brick', "Brick", ""), ('cells', "Cells", ""), 
                   ('checker', "Checker", ""), ('crackle', "Crackle", ""), ('cylindrical', "Cylindrical", ""),
                   ('gradient', "Gradient", ""), ('function', "Function", ""), 
                   ('granite', "Granite", ""), ('hexagon', "Hexagon", ""), ('julia', "Julia", ""), 
                   ('leopard', "Leopard", ""), ('magnet', "Magnet", ""), ('mandel', "Mandelbrot", ""), 
                   ('marble', "Marble", ""), ('onion', "Onion", ""), ('pavement', "Pavement", ""), ('planar', "Planar", ""), 
                   ('quilted', "Quilted", ""), ('ripples', "Ripples", ""),  ('radial', "Radial", ""),
                   ('spherical', "Spherical", ""),('spiral1', "Spiral1", ""), ('spiral2', "Spiral2", ""), ('spotted', "Spotted", ""), 
                   ('square', "Square", ""),('tiling', "Tiling", ""),('triangular', "Triangular", ""),
                   ('waves', "Waves", ""), ('wood', "Wood", ""),('wrinkles', "Wrinkles", "")),
            default='agate')

    magnet_style = EnumProperty(
            name="Magnet style",
            description="magnet or julia",
            items=(('mandel', "Mandelbrot", ""),('julia', "Julia", "")),                   
            default='julia')

    magnet_type = IntProperty(
            name="Magnet_type",
            description="1 or 2",
            min=1, max=2, default=2)

    warp_types = EnumProperty(
            name="Warp Types",
            description="Select the type of warp",
            items=(('CUBIC', "Cubic", ""), ('CYLINDRICAL', "Cylindrical", ""), ('PLANAR', "Planar", ""), ('SPHERICAL', "Spherical", ""),
                    ('REPEAT','Repeat', ""),('TOROIDAL', "Toroidal", ""),('NONE', "None", "No indentation")),
            default='NONE')

    warp_repeat_x = BoolProperty(
            name="Repeat X",
            description="Repeat X",
            default=False)

    rx_value = FloatProperty(
            name="Value",
            description="",
            min=1.0, max=200.0, default=2.0)

    rx_offset = EnumProperty(
            name="Offset",
            description="Select the offset",
            items=(('x', "X", ""), ('y', "Y", ""), ('z', "Z", "")),
            default='y')

    rx_offset_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=200.0, default=0.0)

    rx_flip_x = BoolProperty(
            name="X",
            description="Flip X",
            default=False)

    rx_flip_y = BoolProperty(
            name="Y",
            description="Flip Y",
            default=False)

    rx_flip_z = BoolProperty(
            name="Z",
            description="Flip Z",
            default=False)

    warp_repeat_y = BoolProperty(
            name="Repeat Y",
            description="Repeat Y",
            default=False)

    ry_value = FloatProperty(
            name="Value",
            description="",
            min=1.0, max=200.0, default=2.0)

    ry_offset = EnumProperty(
            name="Offset",
            description="Select the offset",
            items=(('x', "X", ""), ('y', "Y", ""), ('z', "Z", "")),
            default='y')

    ry_offset_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=200.0, default=0.0)

    ry_flip_x = BoolProperty(
            name="X",
            description="Flip X",
            default=False)

    ry_flip_y = BoolProperty(
            name="Y",
            description="Flip Y",
            default=False)

    ry_flip_z = BoolProperty(
            name="Z",
            description="Flip Z",
            default=False)

    warp_repeat_z = BoolProperty(
            name="Repeat Z",
            description="Repeat Z",
            default=False)

    rz_value = FloatProperty(
            name="Value",
            description="",
            min=1.0, max=200.0, default=2.0)

    rz_offset = EnumProperty(
            name="Offset",
            description="Select the offset",
            items=(('x', "X", ""), ('y', "Y", ""), ('z', "Z", "")),
            default='y')

    rz_offset_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=200.0, default=0.0)

    rz_flip_x = BoolProperty(
            name="X",
            description="Flip X",
            default=False)

    rz_flip_y = BoolProperty(
            name="Y",
            description="Flip Y",
            default=False)

    rz_flip_z = BoolProperty(
            name="Z",
            description="Flip Z",
            default=False)

    warp_orientation = EnumProperty(
            name="Warp Orientation",
            description="Select the orientation of warp",
            items=(('x', "X", ""), ('y', "Y", ""), ('z', "Z", "")),
            default='y')

    warp_dist_exp = FloatProperty(
            name="Distance exponent",
            description="Distance exponent",
            min=0.0, max=100.0, default=1.0)
            
    warp_tor_major_radius = FloatProperty(
            name="Major radius",
            description="Torus is distance from major radius",
            min=0.0, max=5.0, default=1.0)            


    warp_turbulence_x = FloatProperty(
            name="Turbulence X",
            description="Turbulence X",
            min=0.0, max=5.0, default=0.0)
            
    warp_turbulence_y = FloatProperty(
            name="Turbulence Y",
            description="Turbulence Y",
            min=0.0, max=5.0, default=0.0) 

    warp_turbulence_z = FloatProperty(
            name="Turbulence Z",
            description="Turbulence Z",
            min=0.0, max=5.0, default=0.0)
            
    modifier_octaves = IntProperty(
            name="Turbulence octaves",
            description="Turbulence octaves",
            min=1, max=10, default=1)            

    modifier_lambda = FloatProperty(
            name="Turbulence lambda",
            description="Turbulence lambda",
            min=0.0, max=5.0, default=1.00)
            
    modifier_omega = FloatProperty(
            name="Turbulence omega",
            description="Turbulence omega",
            min=0.0, max=10.0, default=1.00) 

    modifier_phase = FloatProperty(
            name="Phase",
            description="The phase value causes the map entries to be shifted so that the map starts and ends at a different place.",
            min=0.0, max=2.0, default=0.0)
            
    modifier_frequency = FloatProperty(
            name="Frequency",
            description="The frequency keyword adjusts the number of times that a color map repeats over one cycle of a pattern.",
            min=0.0, max=25.0, default=2.0) 
            
    modifier_numbers = IntProperty(
            name="Numbers",
            description="Numbers",
            min=1, max=27, default=2)             

    modifier_control0 = IntProperty(
            name="Control0",
            description="Control0",
            min=0, max=100, default=1)
            
    modifier_control1 = IntProperty(
            name="Control1",
            description="Control1",
            min=0, max=100, default=1)

    agate_turbulence = FloatProperty(
            name="Turbulence",
            description="Agate turbulence",
            min=0.0, max=5.0, default=2.0) 

    brick_size_x = FloatProperty(
            name="Brick size x",
            description="",
            min=0.0000, max=1.0000, default=0.2500) 

    brick_size_y = FloatProperty(
            name="Brick size y",
            description="",
            min=0.0000, max=1.0000, default=0.0525)
            
    brick_size_z = FloatProperty(
            name="Brick size z",
            description="",
            min=0.0000, max=1.0000, default=0.1250) 

    brick_mortar = FloatProperty(
            name="Mortar",
            description="Mortar",
            min=0.000, max=1.500, default=0.01)

    crackle_form_x = FloatProperty(
            name="X",
            description="Form vector X",
            min=-150.0, max=150.0, default=-1)

    crackle_form_y = FloatProperty(
            name="Y",
            description="Form vector Y",
            min=-150.0, max=150.0, default=1)

    crackle_form_z = FloatProperty(
            name="Z",
            description="Form vector Z",
            min=-150.0, max=150.0, default=0)

    crackle_metric = FloatProperty(
            name="Metric",
            description="Crackle metric",
            min=0.0, max=150.0, default=1)

    crackle_solid = BoolProperty(
            name="Solid",
            description="Crackle solid",
            default=False)

    julia_complex_1 = FloatProperty(
            name="Julia Complex 1",
            description="",
            min=0.000, max=1.500, default=0.360)

    julia_complex_2 = FloatProperty(
            name="Julia Complex 2",
            description="",
            min=0.000, max=1.500, default=0.250)

    f_iter = IntProperty(
            name="Fractal Iteration",
            description="",
            min=0, max=100, default=20)

    f_exponent = IntProperty(
            name="Fractal Exponent",
            description="",
            min=2, max=33, default=2)

    f_ior = IntProperty(
            name="Fractal Interior",
            description="",
            min=1, max=6, default=1)

    f_ior_fac = FloatProperty(
            name="Fractal Interior Factor",
            description="",
            min=0.0, max=10.0, default=1.0)

    f_eor = IntProperty(
            name="Fractal Exterior",
            description="",
            min=1, max=8, default=1)

    f_eor_fac = FloatProperty(
            name="Fractal Exterior Factor",
            description="",
            min=0.0, max=10.0, default=1.0)

    grad_orient_x= IntProperty(
            name="Gradient orientation X",
            description="",
            min=0, max=1, default=0)

    grad_orient_y= IntProperty(
            name="Gradient orientation Y",
            description="",
            min=0, max=1, default=1)

    grad_orient_z= IntProperty(
            name="Gradient orientation Z",
            description="",
            min=0, max=1, default=0)

    pave_sides = EnumProperty(
            name="Pavement sides",
            description="",
            items=(('3', "3", ""), ('4', "4", ""), ('6', "6", "")),
            default='3')

    pave_pat_2= IntProperty(
            name="Pavement pattern 2",
            description="maximum: 2",
            min=1, max=2, default=2)

    pave_pat_3= IntProperty(
            name="Pavement pattern 3",
            description="maximum: 3",
            min=1, max=3, default=3)

    pave_pat_4= IntProperty(
            name="Pavement pattern 4",
            description="maximum: 4",
            min=1, max=4, default=4)

    pave_pat_5= IntProperty(
            name="Pavement pattern 5",
            description="maximum: 5",
            min=1, max=5, default=5)

    pave_pat_7= IntProperty(
            name="Pavement pattern 7",
            description="maximum: 7",
            min=1, max=7, default=7)

    pave_pat_12= IntProperty(
            name="Pavement pattern 12",
            description="maximum: 12",
            min=1, max=12, default=12)

    pave_pat_22= IntProperty(
            name="Pavement pattern 22",
            description="maximum: 22",
            min=1, max=22, default=22)

    pave_pat_35= IntProperty(
            name="Pavement pattern 35",
            description="maximum: 35",
            min=1, max=35, default=35)
            
    pave_tiles= IntProperty(
            name="Pavement tiles",
            description="If sides = 6, maximum tiles 5!!!",
            min=1, max=6, default=1)

    pave_form= IntProperty(
            name="Pavement form",
            description="",
            min=0, max=4, default=0)

    wave_type = EnumProperty(
            name="Waves type",
            description="Select the type of waves",
            items=(('ramp', "Ramp", ""), ('sine', "Sine", ""), ('scallop', "Scallop", ""), ('cubic', "Cubic", ""),
                   ('poly', "Poly", ""), ('triangle', 'Triangle', "")),
            default='ramp')            

    gen_noise = IntProperty(
            name="Noise Generators",
            description="Noise Generators",
            min=1, max=3, default=1) 
    #########FUNCTIONS#########################################################################################################################
    #########FUNCTIONS#########################################################################################################################
    
    func_list = EnumProperty(
            name="Functions",
            description="Select the function for create pattern",
            items=(('NONE', "None", "No indentation"),
                   ("f_algbr_cyl1","Algbr cyl1",""), ("f_algbr_cyl2","Algbr cyl2",""),
                   ("f_algbr_cyl3","Algbr cyl3",""), ("f_algbr_cyl4","Algbr cyl4",""),
                   ("f_bicorn","Bicorn",""), ("f_bifolia","Bifolia",""),
                   ("f_blob","Blob",""), ("f_blob2","Blob2",""),
                   ("f_boy_surface","Boy surface",""), ("f_comma","Comma",""),
                   ("f_cross_ellipsoids","Cross ellipsoids",""), ("f_crossed_trough","Crossed trough",""),
                   ("f_cubic_saddle","Cubic saddle",""), ("f_cushion","Cushion",""),
                   ("f_devils_curve","Devils curve",""), ("f_devils_curve_2d","Devils curve 2d",""),
                   ("f_dupin_cyclid","Dupin cyclid",""), ("f_ellipsoid","Ellipsoid",""),
                   ("f_enneper","Enneper",""), ("f_flange_cover","Flange cover",""),
                   ("f_folium_surface","Folium surface",""), ("f_folium_surface_2d","Folium surface 2d",""),
                   ("f_glob","Glob",""), ("f_heart","Heart",""),
                   ("f_helical_torus","Helical torus",""), ("f_helix1","Helix1",""),
                   ("f_helix2","Helix2",""), ("f_hex_x","Hex x",""),
                   ("f_hex_y","Hex y",""), ("f_hetero_mf","Hetero mf",""),
                   ("f_hunt_surface","Hunt surface",""), ("f_hyperbolic_torus","Hyperbolic torus",""),
                   ("f_isect_ellipsoids","Isect ellipsoids",""), ("f_kampyle_of_eudoxus","Kampyle of eudoxus",""),
                   ("f_kampyle_of_eudoxus_2d","Kampyle of eudoxus 2d",""), ("f_klein_bottle","Klein bottle",""),
                   ("f_kummer_surface_v1","Kummer surface v1",""), ("f_kummer_surface_v2","Kummer surface v2",""),
                   ("f_lemniscate_of_gerono","Lemniscate of gerono",""), ("f_lemniscate_of_gerono_2d","Lemniscate of gerono 2d",""),
                   ("f_mesh1","Mesh1",""), ("f_mitre","Mitre",""),
                   ("f_nodal_cubic","Nodal cubic",""), ("f_noise3d","Noise3d",""),
                   ("f_noise_generator","Noise generator",""), ("f_odd","Odd",""),
                   ("f_ovals_of_cassini","Ovals of cassini",""), ("f_paraboloid","Paraboloid",""),
                   ("f_parabolic_torus","Parabolic torus",""), ("f_ph","Ph",""),
                   ("f_pillow","Pillow",""), ("f_piriform","Piriform",""),
                   ("f_piriform_2d","Piriform 2d",""), ("f_poly4","Poly4",""),
                   ("f_polytubes","Polytubes",""), ("f_quantum","Quantum",""),
                   ("f_quartic_paraboloid","Quartic paraboloid",""), ("f_quartic_saddle","Quartic saddle",""),
                   ("f_quartic_cylinder","Quartic cylinder",""), ("f_r","R",""),
                   ("f_ridge","Ridge",""), ("f_ridged_mf","Ridged mf",""),
                   ("f_rounded_box","Rounded box",""), ("f_sphere","Sphere",""),
                   ("f_spikes","Spikes",""), ("f_spikes_2d","Spikes 2d",""),
                   ("f_spiral","Spiral",""), ("f_steiners_roman","Steiners roman",""),
                   ("f_strophoid","Strophoid",""), ("f_strophoid_2d","Strophoid 2d",""),
                   ("f_superellipsoid","Superellipsoid",""), ("f_th","Th",""),
                   ("f_torus","Torus",""), ("f_torus2","Torus2",""),
                   ("f_torus_gumdrop","Torus gumdrop",""), ("f_umbrella","Umbrella",""),
                   ("f_witch_of_agnesi","Witch of agnesi",""), ("f_witch_of_agnesi_2d","Witch of agnesi 2d","")),
                  
            default='NONE')

    func_x = FloatProperty(
            name="FX",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_plus_x = EnumProperty(
            name="Func plus x",
            description="",
            items=(('NONE', "None", ""), ('increase', "*", ""), ('plus', "+", "")),
            default='NONE')

    func_y = FloatProperty(
            name="FY",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_plus_y = EnumProperty(
            name="Func plus y",
            description="",
            items=(('NONE', "None", ""), ('increase', "*", ""), ('plus', "+", "")),
            default='NONE')

    func_z = FloatProperty(
            name="FZ",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_plus_z = EnumProperty(
            name="Func plus z",
            description="",
            items=(('NONE', "None", ""), ('increase', "*", ""), ('plus', "+", "")),
            default='NONE')

    func_P0 = FloatProperty(
            name="P0",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P1 = FloatProperty(
            name="P1",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P2 = FloatProperty(
            name="P2",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P3 = FloatProperty(
            name="P3",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P4 = FloatProperty(
            name="P4",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P5 = FloatProperty(
            name="P5",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P6 = FloatProperty(
            name="P6",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P7 = FloatProperty(
            name="P7",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P8 = FloatProperty(
            name="P8",
            description="",
            min=0.0, max=25.0, default=1.0)

    func_P9 = FloatProperty(
            name="P9",
            description="",
            min=0.0, max=25.0, default=1.0)


    ############################################################################################################################################LH
    ##########Custom texture gamma###############################################################################################################
    tex_gamma_enable = BoolProperty(
            name="Enable custom texture gamma",
            description="Notify some custom gamma for which texture has been precorrected "
                        "without the file format carrying it and only if it differs from your "
                        "OS expected standard (see pov doc)",
            default=False)

    tex_gamma_value = FloatProperty(
            name="Custom texture gamma",
            description="value for which the file was issued e.g. a Raw photo is gamma 1.0",
            min=0.45, max=5.00, soft_min=1.00, soft_max=2.50, default=1.00)

    ##################################CustomPOV Code############################
    #Only DUMMIES below for now:
    replacement_text = StringProperty(
            name="Declared name:",
            description="Type the declared name in custom POV code or an external .inc "
                        "it points at. pigment {} expected",
            default="")

    ############################# CAIRO Image #####################################
    cr_create_tex = BoolProperty(
            name="Create texture with Cairo",
            description="Create texture with Cairo",
            default=False)




###############################################################################
# Object POV properties.
###############################################################################
class RenderPovSettingsObject(PropertyGroup):

    object_write_type = EnumProperty(
            items= [("BOX", "Box", ""),("BOOLEAN", "Boolean object", ""),("CAIRO", "Cairo object", ""),("CONE", "Cone", ""),("CYLINDER", "Cylinder", ""),
                    ("HEIGHTFIELD", "Height Field",""),("LATHE", "Lathe", ""),("MESH", "Mesh", ""),
                    ("PLANE", "Plane", ""),("PRISM", "Prism", ""),
                    ("SMB", "Scattering media box", ""),("SPHERE", "Sphere", ""),
                    ("SPHERESWEEP", "Sphere sweep", ""),("SUPERELLIPSOID", "Superellipsoid", ""),("SUPERTORUS", "Supertorus", ""),("TORUS", "Torus", ""),
                    ("ROUNDPYRAMID", "Round pyramid", "")
                    ],    #("SOR", "Sor", ""),
            name="Write as",
            description="use blender object as",
            default="MESH")

    boolean_data = StringProperty(
            name="Link to:",
            description="Сhoose the original for the link to it")

    boolean_parent = BoolProperty(
            name="Original",
            description="Not use if object is copy from the original!!!",
            default=False)

    povobject_visible = BoolProperty(
            name="Visible",
            description="Povray object for render",
            default=True)

    shape_se_param1 = FloatProperty(
            name="Parameter 1",
            description="",
            min=0.00, max=10.0, default=0.04)

    shape_se_param2 = FloatProperty(
            name="Parameter 2",
            description="",
            min=0.00, max=10.0, default=0.04)

    shape_rp_sides = IntProperty(
            name="Sides", description="Number of side faces",
            min=3, max=128, default=4)

    shape_rp_border = FloatProperty(
            name="Radius",
            description="Border radius",
            min=0.00, max=2.0, default=0.2)

    shape_rp_filled = BoolProperty(
            name="Filled / Wire",
            description="",
            default=True)

    shape_rp_merge = BoolProperty(
            name="Union / Merge",
            description="",
            default=True)

    shape_hf_image = StringProperty(
            name="Image",
            description="Path to image for height field",
            update=call_tex_image_open,
            maxlen=1024, subtype="FILE_PATH") 

    shape_hf_gamma = FloatProperty(
            name="Gamma",
            description="Gamma",
            min=0.0001, max=20.0, default=1.0)

    shape_hf_premultiplied = BoolProperty(
            name="Premultiplied",
            description="Premultiplied",
            default=True)

    shape_hf_smooth = BoolProperty(
            name="Smooth",
            description="Smooth",
            default=False)

    shape_hf_water = FloatProperty(
            name="Level",
            description="Wather level",
            min=0.00, max=1.00, default=0.0)

    shape_hf_hierarchy = BoolProperty(
            name="Hierarchy",
            description="Height field hierarchy",
            default=True)

    shape_ss_type = EnumProperty(
            items= [("linear_spline", "Linear", ""),("cubic_spline", "Cubic", ""),("b_spline", "B-spline", "")],
            name="Spline",
            description="Spline type",
            default="b_spline")

    lathe_spline_type= EnumProperty(
            items= [("bezier_spline", "Bezier", ""),("cubic_spline", "Cubic", ""),("linear_spline", "Linear", ""),("quadratic_spline", "Quadratic", "")],
            name="Spline",
            description="Spline type",
            default="bezier_spline")

    shape_st_radius_major = FloatProperty(
            name="Major radius",
            description="Major radius",
            min=0.00, max=100.00, default=1.0)

    shape_st_radius_minor = FloatProperty(
            name="Minor radius",
            description="Minor radius",
            min=0.00, max=100.00, default=0.25)

    shape_st_ring = FloatProperty(
            name="Ring",
            description="Ring manipulator",
            min=0.0001, max=100.00, default=1.00)

    shape_st_cross = FloatProperty(
            name="Cross",
            description="Cross manipulator",
            min=0.0001, max=100.00, default=1.00)
    ##########################SCATTERING_BOX##################################
    smb_type = IntProperty(
            name="Scattering type",
            description="1 ... 5",
            min=1, max=5, default=5)

    smb_color = FloatVectorProperty(
            name="Scattering Media Color", description="The Scattering media color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.001, 0.001, 0.001), options={'ANIMATABLE'}, subtype='COLOR')

    smb_ext = FloatProperty(
            name="Extinction",
            description="Extinction",
            min=-10.0, max=10.0, default=0.0)

    smb_ecc = FloatProperty(
            name="Eccentricity",
            description="Eccentricity",
            min=-1.0, max=1.0, default=0.0)

    smb_samples = IntProperty(
            name="Scattering samples",
            description="1 ... 256",
            min=1, max=256, default=30)

    ########################## Cairo object ##################################
    cr_composition = BoolProperty(
            name="Create colored composition with Cairo",
            description="Create colored composition with Cairo",
            default=False)

    cr_fill = BoolProperty(
            name="Fill",
            description="Fill",
            default=True)

    cr_stroke = BoolProperty(
            name="Stroke",
            description="Stroke",
            default=False)

    cr_cap_style = EnumProperty(
            items= [("BUTT", "Butt", ""),("ROUND", "Round", ""), ("SQUARE", "Square", "")],
            name="Cap",
            description="Line cap style",
            default="BUTT")

    cr_join_style = EnumProperty(
            items= [("MITER", "Miter", ""),("ROUND", "Round", ""), ("BEVEL", "Bevel", "")],
            name="Join",
            description="Line join style",
            default="MITER")

    cr_line_width = FloatProperty(
            name="Line width",
            description="",
            min=0.00, max=50.0, default=4.0)

    cr_fill_color = FloatVectorProperty(
            name="Fill color", description="Fill color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(1.0, 1.0, 1.0), options={'ANIMATABLE'}, subtype='COLOR')

    cr_fill_color_alpha = FloatProperty(
            name="Fill_color_alpha",
            description="",
            min=0.0, max=1.0, default=1.0)

    cr_stroke_color = FloatVectorProperty(
            name="Stroke color", description="Stroke color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.5, 0.5, 0.5), options={'ANIMATABLE'}, subtype='COLOR')

    cr_stroke_color_alpha = FloatProperty(
            name="Stroke_color_alpha",
            description="",
            min=0.0, max=1.0, default=1.0)

    ######################### Cairo image ###########################################################

    cr_create_image = BoolProperty(
            name="Create image map with Cairo",
            description="Create image map with Cairo",
            default=False)

    cr_image_width = IntProperty(
            name="Width",
            description="Image width",
            min=1, max=25000, default=640)

    cr_image_height = IntProperty(
            name="Height",
            description="Image height",
            min=1, max=25000, default=640)

    cr_image_bg = EnumProperty(
            items=[("ALPHA", "Alpha", "Use alpha"),("COLOR", "Color", "Use color")],
            name="Background",
            description="Color or alpha",
            default="COLOR")

    cr_image_bg_color = FloatVectorProperty(
            name="Background color", description="Background color",
            precision=4, step=0.01, min=0, soft_max=1,
            default=(0.0, 0.0, 0.0), options={'ANIMATABLE'}, subtype='COLOR')

    cr_image_bg_alpha = FloatProperty(
            name="Alpha",
            description="Alpha",
            min=0.0, max=1.0, default=1.0)

    ####################### Nodes ##############################################################################################


    ###### Materials patterns ######################################################################################################
    mat_patterns = EnumProperty(
            items=[("None", "None", ""),
                   ("cubic", "Cubic", ""),
                   ("image_pattern", "Image pattern", ""),
                   ("pigment_pattern", "Pigment pattern", "")],
            name="Pattern",
            description="Compositions of materials of object",
            update=call_mat_pattern_create,
            default="None")            

    pattern_img = StringProperty(
            name="Pigment pattern image",
            description="Path pigment pattern image",
            update=call_tex_image_open,
            maxlen=1024, subtype="FILE_PATH") 

    pp_value1 = FloatProperty(
            name="Value 1",
            description="Value texture 1",
            min=0.0, max=1.0, default=0.0)

    pp_value2 = FloatProperty(
            name="Value 2",
            description="Value texture 2",
            min=0.0, max=1.0, default=1.0)            

    ip_use_alpha = BoolProperty(
            name="Use alpha",
            description="",
            default=False)

    ip_gamma = FloatProperty(
            name="Gamma",
            description="",
            min=0.1, max=10.00, default=0.1)


    refraction_caustics = BoolProperty(
            name="Refraction Caustics",
            description="use only fake refractive caustics (default) or photon based "
                        "reflective/refractive caustics",
            default=True)


    fake_caustics = BoolProperty(
            name="Fake Caustics", description="use only (Fast) fake refractive caustics",
            default=True)

    fake_caustics_power = FloatProperty(
            name="Fake caustics power",
            description="Values typically range from 0.0 to 1.0 or higher. Zero is no caustics. "
                        "Low, non-zero values give broad hot-spots while higher values give "
                        "tighter, smaller simulated focal points",
            min=0.00, max=10.0, soft_min=0.00, soft_max=1.10, default=0.1)

    photons_dispersion = FloatProperty(
            name="Chromatic Dispersion",
            description="Light passing through will be separated according to wavelength. "
                        "This ratio of refractive indices for violet to red controls how much "
                        "the colors are spread out 1 = no dispersion, good values are 1.01 to 1.1",
            min=1.0000, max=10.000, soft_min=1.0000, soft_max=1.1000, precision=4, default=1.0000)

    photons_dispersion_samples = IntProperty(
            name="Dispersion Samples", description="Number of color-steps for dispersion",
            min=2, max=128, default=7)

    reflection_photons = BoolProperty(
            name="Reflective Caustics",
            description="Use this to make your Sauron's ring ;-P",
            default=False)
            
    refraction_photons = BoolProperty(
            name="Photons (on) / Fake (off)", description="",
            default=True)

    m_ior = FloatProperty(
            name="IOR",
            description="Index of refraction",
            update=call_obj_ior,
            min=1.0, max=5.0, default=1.0)

    fade_color = FloatVectorProperty(
            name="Fade color", description="Color of filtered attenuation for transparent materials",
            precision=4, step=0.01, min=0.0, soft_max=1.0,
            default=(0, 0, 0), options={'ANIMATABLE'}, subtype='COLOR')
            
    fade_power = FloatProperty(
            name="Fade power",
            description="",
            min=0, max=100.00, default=0)

    fade_distance = FloatProperty(
            name="Fade distance",
            description="",
            min=0, max=100.00, default=0)

    # Collect photons
    collect_photons = BoolProperty(
            name="Receive Photon Caustics",
            description="Enable object to collect photons from other objects caustics. Turn "
                        "off for objects that don't really need to receive caustics (e.g. objects"
                        " that generate caustics often don't need to show any on themselves)",
            default=True)
    target_photons = BoolProperty(
            name="Target",
            description="",
            default=False)


    pass_through = BoolProperty(
            name="Pass through",
            description="",
            default=False)

    # Importance sampling
    importance_value = FloatProperty(
            name="Radiosity Importance",
            description="Priority value relative to other objects for sampling radiosity rays. "
                        "Increase to get more radiosity rays at comparatively small yet "
                        "bright objects",
            min=0.01, max=1.00, default=1.00)



    ###### BOOLEAN #######################################################################################################            
    povray_boolean = BoolProperty(
            name="Boolean",
            description="",
            default=False)

    boolean_operation_1 = EnumProperty(
            items=[("difference", "Difference", ""),
                   ("intersection", "Intersection", ""),
                   ("merge", "Union", "")],
            name="Operation 1",
            description="",
            default="merge")

    boolean_group_1 = StringProperty(
            name="Group:",
            description="Select group for boolean operations")
            
    boolean_object_1 = StringProperty(
            name="Object:",
            description="Select object for boolean operations")
                                    
    boolean_operation_2 = EnumProperty(
            items=[("difference", "Difference", ""),
                   ("intersection", "Intersection", ""),
                   ("merge", "Union", "")],
            name="Operation 1",
            description="",
            default="difference")

    boolean_group_2 = StringProperty(
            name="Group:",
            description="Select group for boolean operations")
            
    boolean_object_2 = StringProperty(
            name="Object:",
            description="Select object for boolean operations")

    boolean_operation_3 = EnumProperty(
            items=[("difference", "Difference", ""),
                   ("intersection", "Intersection", ""),
                   ("merge", "Union", "")],
            name="Operation 1",
            description="",
            default="intersection")

    boolean_group_3 = StringProperty(
            name="Group:",
            description="Select group for boolean operations")
            
    boolean_object_3 = StringProperty(
            name="Object:",
            description="Select object for boolean operations")


    ############################ object modifiers ##################################
    noshadow = BoolProperty(
            name="No shadow",
            description="Disable object shadow",
            default=False)

    noreflection = BoolProperty(
            name="No reflection",
            description="Disable object reflecrion",
            default=False)

    noimage = BoolProperty(
            name="No image",
            description="Disable object in render",
            default=False)

    ##################################CustomPOV Code############################
    # Only DUMMIES below for now:
    replacement_text = StringProperty(
            name="Declared name:",
            description="Type the declared name in custom POV code or an external .inc "
                        "it points at. Any POV shape expected e.g: isosurface {}",
            default="")

    number_copy = IntProperty(
            name="Number", description="",
            min=1, max=10000, default=1)
            
    copy_around = BoolProperty(
            name="Copy around", description="Copy around",
            default=False)

    around_axis = EnumProperty(
            items=[("x", "X", ""),
                   ("y", "Y", ""),
                   ("z", "Z", "")],
            name="Axis",
            description="",
            default="z")

    copy_distance = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=10000.0, default=1.0)



###############################################################################
# World POV properties.
###############################################################################
class RenderPovSettingsWorld(PropertyGroup):
    ambient_enable = BoolProperty(
            name="Ambient enable", description="Enable ambient light ",
            default=False)

    fog_use = BoolProperty(
            name="Mist", description="Use mist ",
            update=call_fog_use,
            default=False)

    fog_layer = IntProperty(
            name="Layer",
            description="",
            min=1, max=100, default=0)


###############################################################################
# Camera POV properties.
###############################################################################
class RenderPovSettingsCamera(PropertyGroup):
    cam_type = EnumProperty(
            items=[("perspective", "Perspective", ""),
                   ("orthographic", "Orthographic", ""),
                   ("mesh_camera", "Mesh camera", ""),
                   ("fisheye", "Fisheye", ""),
                   ("ultra_wide_angle", "Ultra wide angle", ""),
                   ("omnimax", "Omnimax", ""),
                   ("panoramic", "Panoramic", ""),
                   ("cylindrical", "Cylindrical", ""),
                   ("spherical", "Spherical", "")],
            name="Cameras type",
            description="",
            default="perspective")

    cam_architectural=BoolProperty(
            name="Architectural",
            description="The Architectural Perspective Camera",
            default=True)

    correction_aspect_ratio = FloatProperty(
            name="Value",
            description="Manual correction of aspect",
            min=-1.00, max=1.00, default=0.17)

    cylinder_type = EnumProperty(
            items=[("1", "1", "vertical cylinder, fixed viewpoint"),
                   ("2", "2", "horizontal cylinder, fixed viewpoint"),
                   ("3", "3", "vertical cylinder, viewpoint moves along the cylinder's axis"),
                   ("4", "4", "horizontal cylinder, viewpoint moves along the cylinder's axis")],
            name="Cylindrical type",
            description="",
            default="1")

    rays_per_pixel = FloatProperty(
            name="Rays Per Pixel",
            description="This float parameter controls the number of rays"
                        "that will be shot for each pixel in the output image. "
                        "Each distribution allows different values, but the minimum is always 1.",
            min=1, max=100, default=1)

    mesh_cam_distr = IntProperty(
            name="Distribution", description="",
            min=0, max=3, default=3)

    mesh_cam_obj = StringProperty(
            name="Object:",
            description="Select object for focus")

    #DOF Toggle

    dof_enable = BoolProperty(
            name="Depth Of Field", description="EnablePOV-Ray Depth Of Field ",
            default=False)

    # Aperture (Intensity of the Blur)
    dof_aperture = FloatProperty(
            name="Aperture",
            description="Similar to a real camera's aperture effect over focal blur (though not "
                        "in physical units and independant of focal length). "
                        "Increase to get more blur",
            min=0.01, max=1.00, default=0.25)

    # Aperture adaptive sampling
    dof_samples_min = IntProperty(
            name="Samples Min", description="Minimum number of rays to use for each pixel",
            min=1, max=128, default=96)

    dof_samples_max = IntProperty(
            name="Samples Max", description="Maximum number of rays to use for each pixel",
            min=1, max=128, default=128)

    dof_variance = IntProperty(
            name="Variance",
            description="Minimum threshold (fractional value) for adaptive DOF sampling (up "
                        "increases quality and render time). The value for the variance should "
                        "be in the range of the smallest displayable color difference",
            min=1, max=100000, soft_max=10000, default=256)

    dof_confidence = FloatProperty(
            name="Confidence",
            description="Probability to reach the real color value. Larger confidence values "
                        "will lead to more samples, slower traces and better images",
            min=0.01, max=0.99, default=0.90)

    ##################################CustomPOV Code############################
    # Only DUMMIES below for now:
    replacement_text = StringProperty(
            name="Texts in blend file",
            description="Type the declared name in custom POV code or an external .inc "
                        "it points at. camera {} expected",
            default="")

class RenderPovSettingsParticle(PropertyGroup):
    hair_material = StringProperty(
            name="Material",
            description="",
            default="")

    hair_type = EnumProperty(
            items=[("CONE", "Cone", ""),
                   ("SPLINE", "Spline", ""),
                   ("SPHERESWEEP", "Sphere sweep", "")],
            name="Hair type",
            description="",
            default="CONE")

    sweep_tolerance = FloatProperty(
            name="Tolerance",
            description="Tolerance in Povray",
            min=1.0, max=100000000.0, default=1.0)

    thickness_root = FloatProperty(
            name="Root",
            description="Thickness of roots of hair",
            min=0.00001, max=1.00, default=0.001)

    thickness_tip = FloatProperty(
            name="Tip",
            description="Thickness of tip of hair",
            min=0.00001, max=1.00, default=0.001)
###############################################################################
# Text POV properties.
###############################################################################
class RenderPovSettingsText(PropertyGroup):
    custom_code = BoolProperty(
            name="Custom Code",
            description="Add this text at the top of the exported POV-Ray file",
            default=False)


###############################################################################
# Povray Preferences.
###############################################################################
class PovrayPreferences(AddonPreferences):
    bl_idname = __name__

    filepath_povray = StringProperty(
                name="Povray Location",
                description="Path to renderer executable",
                subtype='FILE_PATH',
                )
    def draw(self, context):
        layout = self.layout
        layout.prop(self, "filepath_povray")

#################################Operators#########################################
class RenderPovTexturePreview(Operator):
    bl_idname = "tex.preview_update"
    bl_label = "Update preview"
    def execute(self, context):
        tex=context.texture
        texPrevName=string_strip_hyphen(bpy.path.clean_name(tex.name))+"_prev"
        tex.type="IMAGE"
        if sys.platform[:3] == "win":
            previewDir="C:\\povray\\preview\\"
        else:
            workDir=os.path.dirname(__file__)
            previewDir=('%s/preview/'%workDir)
        ##################### ini ##########################################
        fileIni=open("%sPreview.ini"%previewDir,"w")
        fileIni.write('Version=3.7\n')
        fileIni.write('Input_File_Name="%s/Preview.pov"\n'%previewDir)
        fileIni.write('Output_File_Name="%s/%s.png"\n'%(previewDir,texPrevName))
        fileIni.write('Library_Path="%s"\n'%previewDir)
        fileIni.write('Width=256\n')
        fileIni.write('Height=256\n')
        fileIni.write('Pause_When_Done=0\n')
        fileIni.write('Output_File_Type=N\n')
        fileIni.write('Output_Alpha=1\n')
        fileIni.write('Antialias=on\n')
        fileIni.write('Sampling_Method=2\n')
        fileIni.write('Antialias_Depth=3\n')
        fileIni.close()
        ##################### inc ##########################################
        fileInc=open("%s/Preview.inc"%previewDir,"w")
        tex = bpy.context.object.active_material.active_texture
        pat = tex.pov
        mappingDif = ("translate <%.4g,%.4g,%.4g> scale <%.4g,%.4g,%.4g>" % \
              (pat.tex_mov_x, pat.tex_mov_y, pat.tex_mov_z,
               1.0 / pat.tex_scale_x, 1.0 / pat.tex_scale_y, 1.0 / pat.tex_scale_z))                            
        fileInc.write("#declare Texture_prev = \n")
        fileInc.write("texture {\n")
        fileInc.write("pigment {\n")
        fileInc.write("%s\n"%pat.colored_patterns)
        if pat.colored_patterns == 'agate': 
            fileInc.write("agate_turb %.4g\n"%pat.agate_turbulence)
        if pat.colored_patterns in {'spiral1', 'spiral2', 'tiling'}: 
            fileInc.write("%s\n"%pat.modifier_numbers)
        if pat.colored_patterns == 'quilted': 
            fileInc.write("control0 %s control1 %s\n"%(pat.modifier_control0, pat.modifier_control1))                            
        if pat.colored_patterns == 'mandel': 
            fileInc.write("%s exponent %s \n"%(pat.f_iter, pat.f_exponent))  
        if pat.colored_patterns == 'julia': 
            fileInc.write("<%.4g, %.4g> %s exponent %s \n"%(pat.julia_complex_1, pat.julia_complex_2, pat.f_iter, pat.f_exponent))   
        if pat.colored_patterns == 'magnet' and pat.magnet_style == 'mandel': 
            fileInc.write("%s mandel %s \n"%(pat.magnet_type, pat.f_iter))  
        if pat.colored_patterns == 'magnet' and pat.magnet_style == 'julia':  
            fileInc.write("%s julia <%.4g, %.4g> %s\n"%(pat.magnet_type, pat.julia_complex_1, pat.julia_complex_2, pat.f_iter)) 
        if pat.colored_patterns in {'mandel', 'julia', 'magnet'}:
            fileInc.write("interior %s, %.4g\n"%(pat.f_ior, pat.f_ior_fac)) 
            fileInc.write("exterior %s, %.4g\n"%(pat.f_eor, pat.f_eor_fac))
        if pat.colored_patterns == 'gradient': 
            fileInc.write("<%s, %s, %s> \n"%(pat.grad_orient_x, pat.grad_orient_y, pat.grad_orient_z))
        if pat.colored_patterns == 'pavement':
            numTiles=pat.pave_tiles
            numPattern=1
            if pat.pave_sides == '4' and pat.pave_tiles == 3: 
                 numPattern = pat.pave_pat_2
            if pat.pave_sides == '6' and pat.pave_tiles == 3: 
                numPattern = pat.pave_pat_3
            if pat.pave_sides == '3' and pat.pave_tiles == 4: 
                numPattern = pat.pave_pat_3
            if pat.pave_sides == '3' and pat.pave_tiles == 5: 
                numPattern = pat.pave_pat_4
            if pat.pave_sides == '4' and pat.pave_tiles == 4: 
                numPattern = pat.pave_pat_5
            if pat.pave_sides == '6' and pat.pave_tiles == 4: 
                numPattern = pat.pave_pat_7
            if pat.pave_sides == '4' and pat.pave_tiles == 5: 
                numPattern = pat.pave_pat_12
            if pat.pave_sides == '3' and pat.pave_tiles == 6: 
                numPattern = pat.pave_pat_12
            if pat.pave_sides == '6' and pat.pave_tiles == 5: 
                numPattern = pat.pave_pat_22
            if pat.pave_sides == '4' and pat.pave_tiles == 6: 
                numPattern = pat.pave_pat_35
            if pat.pave_sides == '6' and pat.pave_tiles == 6: 
                numTiles = 5                                
            fileInc.write("number_of_sides %s number_of_tiles %s pattern %s form %s \n"%(pat.pave_sides, numTiles, numPattern, pat.pave_form))
        ################ functions #####################################################################################################
        if pat.colored_patterns == 'function':                 
            fileInc.write("{ %s"%pat.func_list)
            fileInc.write("(x")
            if pat.func_plus_x != "NONE":
                if pat.func_plus_x =='increase':
                    fileInc.write("*")                                    
                if pat.func_plus_x =='plus':
                    fileInc.write("+")
                fileInc.write("%.4g"%pat.func_x)
            fileInc.write(",y")
            if pat.func_plus_y != "NONE":
                if pat.func_plus_y =='increase':
                    fileInc.write("*")                                    
                if pat.func_plus_y =='plus':
                    fileInc.write("+")
                fileInc.write("%.4g"%pat.func_y)
            fileInc.write(",z")
            if pat.func_plus_z != "NONE":
                if pat.func_plus_z =='increase':
                    fileInc.write("*")                                    
                if pat.func_plus_z =='plus':
                    fileInc.write("+")
                fileInc.write("%.4g"%pat.func_z)
            sort = -1
            if pat.func_list in {"f_comma","f_crossed_trough","f_cubic_saddle","f_cushion","f_devils_curve",
                                 "f_enneper","f_glob","f_heart","f_hex_x","f_hex_y","f_hunt_surface",
                                 "f_klein_bottle","f_kummer_surface_v1","f_lemniscate_of_gerono","f_mitre",
                                 "f_nodal_cubic","f_noise_generator","f_odd","f_paraboloid","f_pillow",
                                 "f_piriform","f_quantum","f_quartic_paraboloid","f_quartic_saddle",
                                 "f_sphere","f_steiners_roman","f_torus_gumdrop","f_umbrella"}:
                sort = 0
            if pat.func_list in {"f_bicorn","f_bifolia","f_boy_surface","f_superellipsoid","f_torus"}:
                sort = 1
            if pat.func_list in {"f_ellipsoid","f_folium_surface","f_hyperbolic_torus",
                                 "f_kampyle_of_eudoxus","f_parabolic_torus","f_quartic_cylinder","f_torus2"}:
                sort = 2
            if pat.func_list in {"f_blob2","f_cross_ellipsoids","f_flange_cover","f_isect_ellipsoids",
                                 "f_kummer_surface_v2","f_ovals_of_cassini","f_rounded_box","f_spikes_2d","f_strophoid"}:
                sort = 3
            if pat.func_list in {"f_algbr_cyl1","f_algbr_cyl2","f_algbr_cyl3","f_algbr_cyl4","f_blob","f_mesh1","f_poly4","f_spikes"}:
                sort = 4
            if pat.func_list in {"f_devils_curve_2d","f_dupin_cyclid","f_folium_surface_2d","f_hetero_mf","f_kampyle_of_eudoxus_2d",
                                 "f_lemniscate_of_gerono_2d","f_polytubes","f_ridge","f_ridged_mf","f_spiral","f_witch_of_agnesi"}:
                sort = 5
            if pat.func_list in {"f_helix1","f_helix2","f_piriform_2d","f_strophoid_2d"}:
                sort = 6
            if pat.func_list == "f_helical_torus":
                sort = 7
            if sort > -1:
                fileInc.write(",%.4g"%pat.func_P0)
            if sort > 0:
                fileInc.write(",%.4g"%pat.func_P1)
            if sort > 1:
                fileInc.write(",%.4g"%pat.func_P2)
            if sort > 2:
                fileInc.write(",%.4g"%pat.func_P3)
            if sort > 3:
                fileInc.write(",%.4g"%pat.func_P4)
            if sort > 4:
                fileInc.write(",%.4g"%pat.func_P5)
            if sort > 5:
                fileInc.write(",%.4g"%pat.func_P6)
            if sort > 6:
                fileInc.write(",%.4g"%pat.func_P7)
                fileInc.write(",%.4g"%pat.func_P8)
                fileInc.write(",%.4g"%pat.func_P9)
            fileInc.write(")}\n")
        ############## end functions ###############################################################
        if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:                        
            fileInc.write("color_map {\n")
        numColor=0
        for el in tex.color_ramp.elements:
            numColor+=1
            pos = el.position
            col=el.color
            colR,colG,colB,colA = col[0],col[1],col[2],1-col[3]
            if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
                fileInc.write("[%.4g color rgbf<%.4g,%.4g,%.4g,%.4g>] \n"%(pos,colR,colG,colB,colA))
            if pat.colored_patterns in {'brick','checker'} and numColor < 3:
                fileInc.write("color rgbf<%.4g,%.4g,%.4g,%.4g> \n"%(colR,colG,colB,colA))
            if pat.colored_patterns == 'hexagon' and numColor < 4 :
                fileInc.write("color rgbf<%.4g,%.4g,%.4g,%.4g> \n"%(colR,colG,colB,colA))
            if pat.colored_patterns == 'square' and numColor < 5 :
                fileInc.write("color rgbf<%.4g,%.4g,%.4g,%.4g> \n"%(colR,colG,colB,colA))
            if pat.colored_patterns == 'triangular' and numColor < 7 :
                fileInc.write("color rgbf<%.4g,%.4g,%.4g,%.4g> \n"%(colR,colG,colB,colA))
        if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :                        
            fileInc.write("} \n")                         
        if pat.colored_patterns == 'brick':                        
            fileInc.write("brick_size <%.4g, %.4g, %.4g> mortar %.4g \n"%(pat.brick_size_x, pat.brick_size_y, pat.brick_size_z, pat.brick_mortar))
        fileInc.write("%s \n"%mappingDif) 
        fileInc.write("rotate <%.4g,%.4g,%.4g> \n"%(pat.tex_rot_x, pat.tex_rot_y, pat.tex_rot_z)) 
        fileInc.write("turbulence <%.4g,%.4g,%.4g> \n"%(pat.warp_turbulence_x, pat.warp_turbulence_y, pat.warp_turbulence_z))
        fileInc.write("octaves %s \n"%pat.modifier_octaves)
        fileInc.write("lambda %.4g \n"%pat.modifier_lambda)
        fileInc.write("omega %.4g \n"%pat.modifier_omega)
        fileInc.write("frequency %.4g \n"%pat.modifier_frequency)
        fileInc.write("phase %.4g \n"%pat.modifier_phase)                        
        fileInc.write("}}\n")
        fileInc.close()
        ##################### end write ##########################################

        pov_binary="povray"
        if sys.platform[:3] == "win":
            if platform.architecture()[0]=='32bit':
                pov_binary = "C:\\Program Files\\POV-Ray\\v3.7 RC7\\bin\\pvengine.exe"
            if platform.architecture()[0]=='64bit':
                pov_binary = "C:\\Program Files\\POV-Ray\\v3.7 RC7\\bin\\pvengine64.exe"
        p1=subprocess.Popen(["%s"%pov_binary,"-d","%s/Preview.ini"%previewDir],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
        p1.wait()

        tex.use_nodes = True
        tree = tex.node_tree
        links = tree.links
        for n in tree.nodes:
            tree.nodes.remove(n)
        im = tree.nodes.new("TextureNodeImage")  
        pathPrev="%s/%s.png"%(previewDir,texPrevName)
        im.image = bpy.data.images.load(pathPrev)
        name=pathPrev
        name=name.split("/")
        name=name[len(name)-1]
        im.name = name  
        im.location = 200,200
        previewer = tree.nodes.new('TextureNodeOutput')    
        previewer.label = "Preview"
        previewer.location = 400,400
        links.new(im.outputs[0],previewer.inputs[0])
        tex.extension="CLIP"
        return {'FINISHED'}

    ############################# CAIRO #####################################
class TextureCairoCreate(Operator):
    bl_idname = "tex.cairo_create"
    bl_label = "Create image"

    @classmethod
    def poll(cls, context):
        engine = context.scene.render.engine
        crObs=0
        for obj in bpy.data.objects:
            if obj.pov.object_write_type=="CAIRO":
                crObs+=1
        return crObs>0

    def execute(self, context):
        import subprocess, os, mathutils
        from math import pi
        ob=context.object
        scene = context.scene
        cell="_IP"
        if ob.pov.mat_patterns=="pigment_pattern":
            cell="_PP"
        povObName=string_strip_hyphen(bpy.path.clean_name(ob.name))+cell
        if context.texture:
            tex=context.texture
            povObName=string_strip_hyphen(bpy.path.clean_name(tex.name))
            tex.type="IMAGE"
            pathSave=tex.pov.tex_image_path
        else:
            pathSave=ob.pov.pattern_img
        workDir=os.path.dirname(__file__)
        cairoDir=('%s/cairo'%workDir)
        if sys.platform[:3] == "win":
            cairoDir="C:\\povray\\cairo"
            pathSave=pathSave.replace("\\","\\\\")
        if os.path.exists(cairoDir)==False and pathSave == "":
            os.makedirs('%s'%cairoDir)
        if pathSave == "":
            pathSave = "%s/%s.png"%(cairoDir,povObName)
            if sys.platform[:3] == "win":
                pathSave="%s\\%s.png"%(cairoDir,povObName)
            if context.texture:
                tex.pov.tex_image_path = pathSave
            else:
                ob.pov.pattern_img = pathSave
        w=ob.pov.cr_image_width
        h=ob.pov.cr_image_height
        ds=64
        file=open("%s/cairo_write.py"%cairoDir,"w")
        file.write("#!/usr/bin/python3\n")
        file.write("import cairo\n")
        file.write("surface = cairo.ImageSurface (cairo.FORMAT_ARGB32,%s,%s)\n"%(w, h))
        file.write("cr = cairo.Context(surface)\n")
        #file.write("cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)\n")
        file.write("cr.set_antialias(cairo.ANTIALIAS_GRAY)\n")
        if ob.pov.cr_image_bg=="COLOR":
            color=ob.pov.cr_image_bg_color
            color="%.4g,%.4g,%.4g"%(color[0],color[1],color[2])
            file.write("cr.set_source_rgb(%s)\n"%color)
            file.write("cr.paint()\n")
        file.write("cr.translate(%s,%s)\n"%(w/2,h/2))
        global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
        layers=bpy.context.scene.layers
        for layer in range(0,len(layers)):
            for obj in bpy.data.objects:
                if obj.pov.object_write_type=="CAIRO" and obj.type=="CURVE" and obj.layers[layer]:
                    scene_frames = [scene.frame_current]
                    if obj.pov.cr_composition:
                        scene_frames = range(scene.frame_start, scene.frame_end + 1)
                    for frame in scene_frames:
                        scene.frame_set(frame, 0.0)
                        obMatrix=global_matrix*obj.matrix_world
                        splines=obj.data.splines   #bpy.context.object.data.shape_keys.reference_key.data[0].co
                        val=1-obj.data.shape_keys.key_blocks["Key 1"].value
                        for spl in splines:
                            points=spl.bezier_points
                            refPoints=obj.data.shape_keys.reference_key.data
                            for i in range(0,len(points)+1):
                                if i==0:
                                    coPoint=obMatrix*points[i].co
                                    coRefPoint=obMatrix*refPoints[i].co
                                    difMoveX=(coPoint[0]-coRefPoint[0])
                                    coMoveX=coPoint[0]-(difMoveX*val)
                                    difMoveY=(coPoint[2]-coRefPoint[2])
                                    coMoveY=coPoint[2]-(difMoveY*val)
                                    file.write("cr.move_to(%.4g,%.4g)\n"%(coMoveX*ds,coMoveY*ds))
                                if i>0 and i<len(points):
                                    hr=obMatrix*points[i-1].handle_right
                                    hrRef=obMatrix*refPoints[i-1].handle_right
                                    difHrX=(hr[0]-hrRef[0])
                                    coHrX=hr[0]-(difHrX*val)
                                    difHrY=(hr[2]-hrRef[2])
                                    coHrY=hr[2]-(difHrY*val)
                                    hl=obMatrix*points[i].handle_left
                                    hlRef=obMatrix*refPoints[i].handle_left
                                    difHlX=(hl[0]-hlRef[0])
                                    coHlX=hl[0]-(difHlX*val)
                                    difHlY=(hl[2]-hlRef[2])
                                    coHlY=hl[2]-(difHlY*val)
                                    co=obMatrix*points[i].co
                                    coRef=obMatrix*refPoints[i].co
                                    difCoX=(co[0]-coRef[0])
                                    coCoX=co[0]-(difCoX*val)
                                    difCoY=(co[2]-coRef[2])
                                    coCoY=co[2]-(difCoY*val)
                                    file.write("cr.curve_to(%.4g,%.4g,%.4g,%.4g,%.4g,%.4g)\n"%(coHrX*ds,coHrY*ds,coHlX*ds,coHlY*ds,coCoX*ds,coCoY*ds))
                                if i==len(points):
                                    hr=obMatrix*points[i-1].handle_right
                                    hrRef=obMatrix*refPoints[i-1].handle_right
                                    difHrX=(hr[0]-hrRef[0])
                                    coHrX=hr[0]-(difHrX*val)
                                    difHrY=(hr[2]-hrRef[2])
                                    coHrY=hr[2]-(difHrY*val)
                                    hl=obMatrix*points[0].handle_left
                                    hlRef=obMatrix*refPoints[0].handle_left
                                    difHlX=(hl[0]-hlRef[0])
                                    coHlX=hl[0]-(difHlX*val)
                                    difHlY=(hl[2]-hlRef[2])
                                    coHlY=hl[2]-(difHlY*val)
                                    co=obMatrix*points[0].co
                                    coRef=obMatrix*refPoints[0].co
                                    difCoX=(co[0]-coRef[0])
                                    coCoX=co[0]-(difCoX*val)
                                    difCoY=(co[2]-coRef[2])
                                    coCoY=co[2]-(difCoY*val)
                                    file.write("cr.curve_to(%.4g,%.4g,%.4g,%.4g,%.4g,%.4g)\n"%(coHrX*ds,coHrY*ds,coHlX*ds,coHlY*ds,coCoX*ds,coCoY*ds))
                        #file.write("cr.close_path ()\n")
                        if obj.pov.cr_fill:
                            color=obj.pov.cr_fill_color
                            color="%.4g,%.4g,%.4g,%.4g"%(color[0],color[1],color[2],obj.pov.cr_fill_color_alpha)
                            file.write("cr.set_source_rgba(%s)\n"%color)
                            file.write("cr.fill()\n")
                        if obj.pov.cr_stroke:
                            file.write("cr.set_line_width(%.4g)\n"%obj.pov.cr_line_width)
                            color=obj.pov.cr_stroke_color
                            color="%.4g,%.4g,%.4g,%.4g"%(color[0],color[1],color[2],obj.pov.cr_stroke_color_alpha)
                            file.write("cr.set_source_rgba(%s)\n"%color)
                            file.write("cr.stroke()\n")
        file.write('surface.write_to_png("%s")\n'%pathSave)
        file.close()
        py_binary="python"
        if sys.platform[:3] == "win":
            py_binary="C:\\Python27\python"
        p1=subprocess.Popen(["%s"%py_binary,"%s/cairo_write.py"%cairoDir])
        p1.wait()
        if context.texture:
            tex.use_nodes = True
            tree = tex.node_tree
            links = tree.links
            for n in tree.nodes:
                tree.nodes.remove(n)
            im = tree.nodes.new("TextureNodeImage")  
            im.image = bpy.data.images.load(tex.pov.tex_image_path)
            name=tex.pov.tex_image_path
            name=name.split("/")
            name=name[len(name)-1]
            im.name = name  
            im.location = 200,200
            previewer = tree.nodes.new('TextureNodeOutput')    
            previewer.label = "Preview"
            previewer.location = 400,400
            links.new(im.outputs[0],previewer.inputs[0]) 
            tex.pov.tex_images_data=name        
        return {'FINISHED'}

class ExportHair(Operator):
    bl_idname = "ob.exporthair"
    bl_label = "Add hair"

    def execute(self, context):
        import bmesh
        import mathutils
        from math import pi, modf
        scene=context.scene
        ob=context.object
        global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
        matrix=global_matrix*ob.matrix_world
        for pSys in ob.particle_systems:
            for mod in ob.modifiers:
                if mod.type == 'PARTICLE_SYSTEM':
                    if mod.particle_system.settings.type == 'HAIR' and mod.particle_system.name==pSys.name:
                        modPovName=string_strip_hyphen(bpy.path.clean_name(mod.particle_system.name))
                        numberHairs = (len(mod.particle_system.particles) + len(mod.particle_system.child_particles))
                        file=open("%s/%s.inc"%(scene.pov.scene_path,modPovName),'w')
                        file.write("#declare Array%s = array[%s] {\n"%(modPovName,len(mod.particle_system.child_particles)))
                        radiusRoot = mod.particle_system.settings.pov.thickness_root
                        radiusTip = mod.particle_system.settings.pov.thickness_tip
                        child = mod.particle_system.settings.child_nbr
                        for i in range(len(mod.particle_system.particles),numberHairs):
                            numPt=0
                            for k in range(0,200):
                                co = pSys.co_hair(ob,i,k)
                                if co[0]==co[1]==co[2]==0:
                                    continue
                                else:
                                    numPt+=1
                            numPtAll=numPt+4
                            if mod.particle_system.settings.pov.hair_type == "CONE":
                                coRoot = pSys.co_hair(ob,i,0)
                                coTip = pSys.co_hair(ob,i,numPt-1)
                                file.write("cone{\n")
                                centerRoot = "<%.4g,%.4g,%.4g>"%(coRoot[:])
                                file.write("%s, %.4g,\n"%(centerRoot,radiusRoot))
                                centerTip = "<%.4g,%.4g,%.4g>"%(coTip[:])
                                file.write("%s, %.4g},\n"%(centerTip,radiusTip))
                            if mod.particle_system.settings.pov.hair_type == "SPLINE":
                                file.write("spline { cubic_spline\n")
                                co = pSys.co_hair(ob,i,0)
                                file.write("0/%s <%.4g,%.4g,%.4g>\n"%(numPtAll,co[0],co[1],co[2]))
                                file.write("1/%s <%.4g,%.4g,%.4g>\n"%(numPtAll,co[0],co[1],co[2]))
                                for p in range(0,numPt):
                                    co = pSys.co_hair(ob,i,p)
                                    file.write("%s/%s <%.4g,%.4g,%.4g>\n"%(p+2,numPtAll,co[0],co[1],co[2]))
                                file.write("%s/%s <%.4g,%.4g,%.4g>\n"%(numPtAll-2,numPtAll,co[0],co[1],co[2]))
                                file.write("%s/%s <%.4g,%.4g,%.4g>\n"%(numPtAll-1,numPtAll,co[0],co[1],co[2]))
                                file.write("},\n") 
                            if mod.particle_system.settings.pov.hair_type == "SPHERESWEEP":
                                file.write("sphere_sweep { cubic_spline %s\n"%numPtAll)
                                co = pSys.co_hair(ob,i,0)
                                file.write("<%.4g,%.4g,%.4g>,%.4g\n"%(co[0],co[1],co[2],radiusRoot))
                                file.write("<%.4g,%.4g,%.4g>,%.4g\n"%(co[0],co[1],co[2],radiusRoot))
                                deltaRadius=(radiusRoot-radiusTip)/(numPt-1)
                                radius=radiusRoot+deltaRadius
                                for p in range(0,numPt):
                                    radius-=deltaRadius
                                    co = pSys.co_hair(ob,i,p)
                                    file.write("<%.4g,%.4g,%.4g>,%.4g\n"%(co[0],co[1],co[2],radius))
                                file.write("<%.4g,%.4g,%.4g>,%.4g\n"%(co[0],co[1],co[2],radiusTip))
                                file.write("<%.4g,%.4g,%.4g>,%.4g\n"%(co[0],co[1],co[2],radiusTip))
                                tolerance=mod.particle_system.settings.pov.sweep_tolerance/1000000
                                file.write("tolerance %s\n"%tolerance)
                                file.write("},\n")
                        file.write("}\n")
                        uv_color=False
                        material=None
                        finish=""
                        color="rgb 1"
                        if mod.particle_system.settings.pov.hair_material:
                            for mat in ob.data.materials:
                                if mat.name == mod.particle_system.settings.pov.hair_material:
                                    material=mat
                                    if mat.pov and material.pov.povray_finish:
                                        finish="%s_finish"%string_strip_hyphen(bpy.path.clean_name(material.name))
                                    if ob.data.uv_layers and mat.active_texture and mat.active_texture.pov.tex_images_data:
                                        uv_color=True
                                        file.write("#declare ArrayColors%s = array[%s] {\n"%(modPovName,len(mod.particle_system.child_particles)))
                                        tex = mat.active_texture.pov.tex_images_data
                                        image=bpy.data.images[tex]
                                        image_width = image.size[0]
                                        image_height = image.size[1]
                                        image_pixels = image.pixels[:]
                                        for p in range(0,child):
                                            for i in range(0,len(mod.particle_system.particles)):
                                                uv_co = pSys.uv_on_emitter(mod, pSys.particles[i], i, 0)
                                                x_co = round(uv_co[0] * (image_width - 1))
                                                y_co = round(uv_co[1] * (image_height - 1))
                                                pixelnumber = (image_width * y_co) + x_co
                                                r = image_pixels[pixelnumber*4]
                                                g = image_pixels[pixelnumber*4+1]
                                                b = image_pixels[pixelnumber*4+2]
                                                a = image_pixels[pixelnumber*4+3]
                                                file.write("rgbt <%.4g,%.4g,%.4g,%.4g>\n"%(r,g,b,a))
                                        file.write("}\n")
                        if mod.particle_system.settings.pov.hair_type == "SPLINE":
                            file.write("#declare Segment%s =\n"%modPovName)
                            file.write("sphere {0,%.4g}\n\n"%radiusRoot)
                            file.write("#declare HairStep = 1;\n")
                            file.write("#declare Hairs%s =\n"%modPovName)
                            file.write("union {\n")
                            file.write("    #declare I = 0;\n")
                            file.write("    #while (I < %s)\n"%len(mod.particle_system.child_particles))
                            file.write("        union {\n")
                            file.write("            #declare ctr = 0;\n")
                            file.write("            #declare col = 0;\n")
                            file.write("            #while (ctr < 1)\n")
                            file.write("                object { Segment%s texture{pigment{rgbt col}}\n"%modPovName)
                            file.write("                        translate Array%s[I](ctr)}\n"%modPovName)
                            file.write("            #declare ctr = ctr + 0.0025;\n")
                            file.write("            #if (ctr > 0.5)\n")
                            file.write("                #declare col = col + 0.0025;\n")
                            file.write("                #end\n")
                            file.write("            #if (ctr > 0.6)\n")
                            file.write("                #declare col = 1;\n")
                            file.write("                #end\n")
                            file.write("            #end }\n")
                            file.write("#declare I = I + HairStep ;\n")
                            file.write("#end }\n")
                            file.close()
                        if mod.particle_system.settings.pov.hair_type in {"CONE", "SPHERESWEEP"}:

                            file.write("#declare HairStep = 1;\n")
                            file.write("#declare Hairs%s =\n"%modPovName)
                            file.write("union {\n")
                            file.write("    #declare I = 0;\n")
                            file.write("    #while (I < %s)\n"%len(mod.particle_system.child_particles))
                            if uv_color:
                                file.write("            object { Array%s[I] texture{pigment{ArrayColors%s[I]} finish{%s}}}\n"%(modPovName,modPovName,finish))
                            elif uv_color==False and material.pov and material.pov.finish_as_mat:
                                color="%.4g,%.4g,%.4g"%material.diffuse_color[:]
                                colorft="rgb <%s>"%color
                                if material.pov.m_transparent:
                                    colorft="rgbft <%s,%.4g,%.4g>"%(color,material.pov.m_filter,material.pov.m_transmit)
                                file.write("            object { Array%s[I] texture{pigment{%s} finish{%s}}}\n"%(modPovName,colorft,finish))
                            else:
                                file.write("            object { Array%s[I] texture{pigment{rgb 1} finish{%s}}}\n"%(modPovName,finish))
    
                            file.write("#declare I = I + HairStep ;\n")
                            file.write("#end }\n")
                            file.close()

        return {'FINISHED'}
class ObjectCopy(Operator):
    bl_idname = "ob.objectcopy"
    bl_label = "Create copy"

    def execute(self, context):
        from math import pi,sin,cos
        ob=context.object
        numberCopy=ob.pov.number_copy
        rad=ob.pov.copy_distance
        step=(360/numberCopy)*pi/180
        val=0
        if ob.pov.copy_around:
            val=-step
        for i in range (0,numberCopy):
            if (ob.pov.copy_around and i != numberCopy - 1) or ob.pov.copy_around == False:
                bpy.ops.object.duplicate(linked=False, mode='TRANSLATION')
            if ob.pov.copy_around:
                val+=step
            if ob.pov.copy_around == False:
                val+=rad
            x=val
            y=0
            value=(x, y, 0)
            if ob.pov.copy_around:
                x=rad*sin(val)
                y=rad*cos(val)
                if ob.pov.around_axis=="x":
                    value=(0, x, -y)
                if ob.pov.around_axis=="y":
                    value=(x, 0, y)
                if ob.pov.around_axis=="z":
                    value=(-x, y, 0)
            bpy.ops.transform.translate(value=value)
            if ob.pov.copy_around:
                if ob.pov.around_axis=="x":
                    axis=(1, 0, 0)
                if ob.pov.around_axis=="y":
                    axis=(0, 1, 0)
                if ob.pov.around_axis=="z":
                    axis=(0, 0, 1)
                bpy.ops.transform.rotate(value=val, axis=axis)
            for obj in bpy.data.objects:
                if obj.select==True:
                    if ob.pov.object_write_type=='BOOLEAN' and ob.pov.boolean_parent:
                        obj.pov.boolean_parent=False
                        obj.pov.boolean_data=ob.name
                    obj.select=False
            bpy.context.scene.objects.active=ob
            ob.select=True
        return {'FINISHED'}

class ObjectBooleanMesh(Operator):
    bl_idname = "ob.booleanmesh"
    bl_label = "Create simulation"

    def execute(self, context):
        ob=context.object
        if ob.type=='CURVE'and ob.pov.object_write_type=='LATHE':
            res=ob.data.bevel_resolution
        obName=ob.name
        ob.name=obName+'_boolean'
        first=[]
        second=[]
        third=[]
        bo=ob.pov.boolean_object_1
        bg=ob.pov.boolean_group_1
        bo2=ob.pov.boolean_object_2
        bg2=ob.pov.boolean_group_2
        bo3=ob.pov.boolean_object_3
        bg3=ob.pov.boolean_group_3
        if bo != '':
            first.append(bo)
        if bg != '':
            for group in bpy.data.groups:
                if group.name == bg:
                    for obG in group.objects:
                        first.append(obG.name)
        if bo2 != '':
            second.append(bo2)
        if bg2 != '':
            for group in bpy.data.groups:
                if group.name == bg2:
                    for obG in group.objects:
                        second.append(obG.name)
        if bo3 != '':
            third.append(bo3)
        if bg3 != '':
            for group in bpy.data.groups:
                if group.name == bg3:
                    for obG in group.objects:
                        third.append(obG.name)
        bpy.ops.object.duplicate(linked=False, mode='TRANSLATION')

        for obj in bpy.data.objects:
            if obj.select==True:
                obj.name=obName
                obj.select=False
        bpy.context.scene.objects.active=ob
        ob.select=True
        if (ob.type=='CURVE'and ob.pov.object_write_type != 'LATHE') or ob.type == 'META':
            bpy.ops.object.convert(target='MESH',keep_original=False)
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.object.mode_set(mode='OBJECT')
        if ob.type=='CURVE'and ob.pov.object_write_type=='LATHE':
            from math import pi
            bpy.ops.object.convert(target='MESH',keep_original=False)
            bpy.context.area.type = 'VIEW_3D'
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.spin(steps=res, dupli=False, angle=2*pi, center=(0, 1, 0), axis=(0, 1, 0))
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.remove_doubles(threshold=0.0001, use_unselected=False)
            bpy.ops.mesh.normals_make_consistent(inside=False)
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.context.area.type = 'PROPERTIES'
        ob.pov.object_write_type='BOOLEAN'
        ob.pov.boolean_parent=True
        p_b_o=bpy.context.scene.povray_boolean_objects.add()
        p_b_o.name=ob.name
        dupli=[first,second,third]
        operations=[ob.pov.boolean_operation_1,ob.pov.boolean_operation_2,ob.pov.boolean_operation_3]

        def createBooleanMesh(line,operation):
            for i in range(0,len(line)):
                operation=operation
                obDupli=bpy.data.objects['%s'%line[i]]

                nameB=obDupli.name

                mods = ob.modifiers
                mod = mods.new("Booly",type='BOOLEAN')
                if operation == "merge":
                    mod.operation='UNION'
                elif operation == "intersection":
                    mod.operation='INTERSECT'
                else:
                    mod.operation = operation.upper()
                mod.object=bpy.data.objects['%s'%nameB]


                bpy.ops.object.modifier_apply(apply_as='DATA', modifier="Booly")


        for i in range(0,3):
            line=dupli[i]
            operation=operations[i]
            createBooleanMesh(line,operation)
        ob.pov.boolean_object_1=''
        ob.pov.boolean_group_1=''
        ob.pov.boolean_object_2=''
        ob.pov.boolean_group_2=''
        ob.pov.boolean_object_3=''
        ob.pov.boolean_group_3=''
        ob.pov.povray_boolean=False

        return {'FINISHED'}

class PovrayBooleanObjects(bpy.types.PropertyGroup):
    name = bpy.props.StringProperty(name="Name",default="")

class UpdateBooleanCollection(Operator):
    bl_idname = "ob.update_pbo"
    bl_label = "Update collection"
    def execute(self, context):
        scene=context.scene
        for item in scene.povray_boolean_objects:
            scene.povray_boolean_objects.remove(1)
        for ob in bpy.data.objects:
            if ob.type=='MESH' and ob.pov.object_write_type=='BOOLEAN' and ob.pov.boolean_parent:
                newItem=scene.povray_boolean_objects.add()
                newItem.name=ob.name
        return {'FINISHED'}

class PovraySlopeMapEntry(PropertyGroup):
    slope_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=0.00)
    slope_height = FloatProperty(
            name="Height",
            description="",
            min=0.00, max=1.00, default=0.00)
    slope_slope = IntProperty(
            name="Slope",
            description="",
            min=-3, max=3, default=0)

class SlopeEntryNew(Operator):
    bl_idname = "tex.slopeentrynew"
    bl_label = "New entry"

    def execute(self, context):
        tex=context.texture
        index=-1
        for i in range(0,len(tex.povray_slope)):
            index+=1
        entry=tex.povray_slope.add()
        entry.name="slope_entry.%s"%(index+1)
        entry.slope_value=0
        entry.slope_height=0
        entry.slope_slope=0
        return {'FINISHED'}

class SlopeEntryRemove(Operator):
    bl_idname = "tex.slopeentryremove"
    bl_label = "Remove entry"

    def execute(self, context):
        tex=context.texture
        number=tex.pov.slope_entry_number
        tex.povray_slope.remove(number)
        return {'FINISHED'}

class PovrayTextureMapEntry(PropertyGroup):
    texture_value = FloatProperty(
            name="Value",
            description="",
            min=0.0, max=1.0, default=0.0)
    name = StringProperty(
            name="Texture name",
            description="")

class TextureEntryNew(Operator):
    bl_idname = "tex.textureentrynew"
    bl_label = "New entry"

    def execute(self, context):
        tex=context.texture
        entry=tex.povray_texture.add()
        entry.name=""
        entry.texture_value=0.0
        return {'FINISHED'}

class TextureEntryRemove(Operator):
    bl_idname = "tex.textureentryremove"
    bl_label = "Remove entry"

    def execute(self, context):
        tex=context.texture
        number=tex.pov.pigment_entry_number
        tex.povray_texture.remove(number)
        return {'FINISHED'}

class PovrayPigmentMapEntry(PropertyGroup):
    pigment_number = IntProperty(
            name="Number",
            description="",
            min=0, max=256, default=0)
    pigment_name = StringProperty(
            name="Pigment name",
            description="")
    color_transmit = FloatProperty(
            name="Transmit",
            description="",
            min=0.0, max=1.0, default=0.0) 
class PigmentEntryNew(Operator):
    bl_idname = "tex.pigmententrynew"
    bl_label = "New entry"

    def execute(self, context):
        tex=context.texture
        index=-1
        for i in range(0,len(tex.povray_pigment)):
            index+=1
        entry=tex.povray_pigment.add()
        entry.name="pigment_entry.%s"%(index+1)
        entry.pigment_number=0
        entry.pigment_name=''
        return {'FINISHED'}

class PigmentEntryRemove(Operator):
    bl_idname = "tex.pigmententryremove"
    bl_label = "Remove entry"

    def execute(self, context):
        tex=context.texture
        number=tex.pov.pigment_entry_number
        tex.povray_pigment.remove(number)
        return {'FINISHED'}

class PovrayPatternColorProps(PropertyGroup):

    name = StringProperty(
            name="Color pattern name",
            description="")
    color_pattern_type = EnumProperty(
            name="Type",
            description="Texture type",
            items=(('NONE', "None", ""),('BRICK', "Brick", ""),('CHECKER', "Checker", ""),
            ('HEXAGON', "Hexagon", ""),('IMAGEPATTERN', "Image Pattern", ""),
            ('LAYERED', "Layered", ""),('PIGMENTPATTERN', "Pigment pattern", "")),
            default='NONE')
    pigment_pattern_color = StringProperty(
            name="Pigment pattern",
            description="")
    color_texture_first = StringProperty(
            name="First texture",
            description="")
    color_texture_second = StringProperty(
            name="Second_texture",
            description="")
    color_texture_third = StringProperty(
            name="Third_texture",
            description="")
    color_first_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=0.00)
    color_second_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=1.00)

class PovrayPatternNormalProps(PropertyGroup):

    name = StringProperty(
            name="Normal pattern name",
            description="")
    pigment_pattern_nor = StringProperty(
            name="Pigment pattern",
            description="")
    nor_texture_first = StringProperty(
            name="First texture",
            description="")
    nor_texture_second = StringProperty(
            name="Second_texture",
            description="")
    nor_first_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=0.00)
    nor_second_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=1.00)


class PovrayLayerNormal(PropertyGroup):
    nor_texture_layer = StringProperty(
            name="Texture",
            description="")
    normal_texture_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=1.00)

class PovrayPigments(PropertyGroup):
    name = StringProperty(
            name="Pigment name",
            description="")
    index = IntProperty(
            name="Pigment index",
            description="",
            min=-1,max=1000,default=-1)

class PovrayPigmentPatterns(PropertyGroup):
    name = StringProperty(
            name="Pigment pattern name",
            description="")
    index = IntProperty(
            name="Pigment pattern index",
            description="",
            min=-1,max=1000,default=-1)

class PovrayNormals(PropertyGroup):
    name = StringProperty(
            name="Normal name",
            description="")
    index = IntProperty(
            name="Normal index",
            description="",
            min=-1,max=1000,default=-1)

class PovrayFinishes(PropertyGroup):
    name = StringProperty(
            name="Finish name",
            description="")
    index = IntProperty(
            name="Finish index",
            description="",
            min=-1,max=1000,default=-1)

class PovrayLayeredTexture(PropertyGroup):
    name = StringProperty(
            name="Texture",
            description="")
    index = IntProperty(
            name="Normal index",
            description="",
            min=-1,max=1000,default=-1)
    layer_value = FloatProperty(
            name="Value",
            description="",
            min=0.00, max=1.00, default=1.00)
    layer_pigment = StringProperty(
            name="Texture",
            update=mat_texslots_update,
            description="")
    layer_normal = StringProperty(
            name="Normal",
            description="")
    layer_finish = StringProperty(
            name="Finish",
            description="")

class PovrayFogs(PropertyGroup):
    fog_type = IntProperty(
            name="Type",
            description="",
            min=1,max=2,default=2)
    fog_color = FloatVectorProperty(
            name="Fog color", description="Color of fog",
            precision=4, step=0.01, min=0.0, soft_max=1.0,
            default=(0.5, 0.5, 0.5), options={'ANIMATABLE'}, subtype='COLOR')
    fog_distance = FloatProperty(
            name="Distance",
            description="",
            min=-10000.00, max=10000.00, default=7.00)
    fog_up_x = IntProperty(
            name="Up X",
            description="",
            min=0, max=1, default=0)
    fog_up_y = IntProperty(
            name="Up Y",
            description="",
            min=0, max=1, default=1)
    fog_up_z = FloatProperty(
            name="Up Z",
            description="",
            min=0, max=1, default=0)
    fog_offset = FloatProperty(
            name="Offset",
            description="",
            min=-10000.00, max=10000.00, default=10.0)
    fog_alt = FloatProperty(
            name="Altitude",
            description="",
            min=-10000.00, max=10000.00, default=2.0)
    fog_turb_x = FloatProperty(
            name="Turbulence X",
            description="",
            min=0.00, max=10.00, default=0.00)
    fog_turb_y = FloatProperty(
            name="Turbulence Y",
            description="",
            min=0.00, max=10.00, default=0.00)
    fog_turb_z = FloatProperty(
            name="Turbulence Z",
            description="",
            min=0.00, max=10.00, default=0.00)
    fog_turb_depth = FloatProperty(
            name="Turb depth",
            description="",
            min=0.00, max=10.00, default=0.00)
    fog_lambda = FloatProperty(
            name="Lambda",
            description="",
            min=0.00, max=10.00, default=0.00)
    fog_octaves = IntProperty(
            name="Octaves",
            description="",
            min=1,max=20,default=1)
    fog_omega = FloatProperty(
            name="Omega",
            description="",
            min=0.00, max=10.00, default=0.00) 
    name = StringProperty(
            name="Fog name",
            description="")
    index = IntProperty(
            name="Fog index",
            description="",
            min=-1,max=1000,default=-1)

class TextureLayerNew(Operator):
    bl_idname = "mat.layernew"
    bl_label = "New layer"

    def execute(self, context):
        mat=context.material
        index=-1
        for i in range(0,len(mat.povray_layered_texture)):
            index+=1
        layer=mat.povray_layered_texture.add()
        layer.name="layer_entry.%s"%(index+1)
        return {'FINISHED'}

class TextureLayerRemove(Operator):
    bl_idname = "mat.layerremove"
    bl_label = "Remove"

    def execute(self, context):
        mat=context.material
        number=mat.pov.m_layer_number
        mat.povray_layered_texture.remove(number)
        return {'FINISHED'}

class FogLayerNew(Operator):
    bl_idname = "world.foglayernew"
    bl_label = "New fog layer"

    def execute(self, context):
        world=context.world
        index=-1
        for i in range(0,len(world.povray_fogs)):
            index+=1
        layer=world.povray_fogs.add()
        layer.index=index+1
        return {'FINISHED'}

class FogLayerRemove(Operator):
    bl_idname = "world.foglayerremove"
    bl_label = "Remove"

    def execute(self, context):
        world=context.world
        number=world.pov.fog_layer-1
        world.povray_fogs.remove(number)
        index=-1
        for layer in world.povray_fogs:
            index+=1
            layer.index=index
        return {'FINISHED'}

###############################################################################
# Text POV properties.
###############################################################################
class ColorRampElementsPov(PropertyGroup):
    pigment = StringProperty(
        name="Pigment",
        description="")

class FormatPNG(bpy.types.Operator, ExportHelper):
    bl_idname       = "export_png_format.png";
    bl_label        = "Save as";
    bl_options      = {'PRESET'};
    filename_ext    = ".png";

    def execute(self, context):
        tex=context.texture
        filepath = self.filepath
        tex.pov.tex_image_path = filepath
        return {'FINISHED'}
##################################END##############################################

def register():
    bpy.utils.register_module(__name__)
    bpy.types.Scene.povray_boolean_objects = CollectionProperty(type=PovrayBooleanObjects)
    bpy.types.Scene.povray_pigments = CollectionProperty(type=PovrayPigments)
    bpy.types.Scene.povray_pigment_patterns = CollectionProperty(type=PovrayPigmentPatterns)
    bpy.types.Scene.povray_normals = CollectionProperty(type=PovrayNormals)
    bpy.types.Scene.povray_finishes = CollectionProperty(type=PovrayFinishes)
    bpy.types.World.povray_fogs = CollectionProperty(type=PovrayFogs)
    bpy.types.Texture.povray_slope = CollectionProperty(type=PovraySlopeMapEntry)
    bpy.types.Texture.povray_pigment = CollectionProperty(type=PovrayPigmentMapEntry)
    bpy.types.Texture.povray_texture = CollectionProperty(type=PovrayTextureMapEntry)
    bpy.types.Material.povray_layered_texture = CollectionProperty(type=PovrayLayeredTexture)
    bpy.types.Lamp.pov = PointerProperty(type=RenderPovSettingsLamp)
    bpy.types.Scene.pov = PointerProperty(type=RenderPovSettingsScene)
    bpy.types.Material.pov = PointerProperty(type=RenderPovSettingsMaterial)
    bpy.types.Texture.pov = PointerProperty(type=RenderPovSettingsTexture)
    bpy.types.Object.pov = PointerProperty(type=RenderPovSettingsObject)
    bpy.types.Camera.pov = PointerProperty(type=RenderPovSettingsCamera)
    bpy.types.Text.pov = PointerProperty(type=RenderPovSettingsText)
    bpy.types.World.pov = PointerProperty(type=RenderPovSettingsWorld)
    bpy.types.ParticleSettings.pov = PointerProperty(type=RenderPovSettingsParticle)
    bpy.types.Texture.pigments = CollectionProperty(type=ColorRampElementsPov)

def unregister():
    bpy.utils.unregister_module(__name__)
    del bpy.types.Scene.povray_boolean_objects
    del bpy.types.Texture.povray_slope
    del bpy.types.Material.povray_pattern_color
    del bpy.types.Material.povray_ppatnor
    del bpy.types.Material.povray_laycolor
    del bpy.types.Material.povray_laynor
    del bpy.types.Lamp.pov
    del bpy.types.Scene.pov
    del bpy.types.Material.pov
    del bpy.types.Texture.pov
    del bpy.types.Object.pov
    del bpy.types.Camera.pov
    del bpy.types.Text.pov
    del bpy.types.World.pov

if __name__ == "__main__":
    register()


